# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Tests for the Dataset class.
"""

import numpy as np
import pandas as pd

from pytest import raises

from lenskit.data import Dataset, from_interactions_df
from lenskit.testing import ml_ds, ml_ratings  # noqa: F401


def test_from_ratings_default_names(ml_ratings: pd.DataFrame):
    ratings = ml_ratings
    ds = from_interactions_df(ratings)
    assert ds.item_count == ratings["item_id"].nunique()
    assert ds.user_count == ratings["user_id"].nunique()


def test_from_ratings_nosuffix(ml_ratings: pd.DataFrame):
    ratings = ml_ratings.rename(columns={"user_id": "user", "item_id": "item"})
    ds = from_interactions_df(ratings)
    assert ds.item_count == ratings["item"].nunique()
    assert ds.user_count == ratings["user"].nunique()


def test_from_ratings_names_upper(ml_ratings: pd.DataFrame):
    ratings = ml_ratings.rename(columns={"user_id": "USER", "item_id": "ITEM"})
    ds = from_interactions_df(ratings)
    assert ds.item_count == ratings["ITEM"].nunique()
    assert ds.user_count == ratings["USER"].nunique()


def test_user_id_single(ml_ds: Dataset):
    users = ml_ds.users.index
    assert ml_ds.users.id(0) == users[0]
    assert ml_ds.users.id(ml_ds.user_count - 1) == users[-1]
    assert ml_ds.users.id(50) == users[50]


def test_user_id_many(ml_ds: Dataset):
    users = ml_ds.users.index
    assert np.all(ml_ds.users.ids([1, 5, 23]) == users[[1, 5, 23]])
    assert np.all(ml_ds.users.ids(np.array([1, 5, 23])) == users[[1, 5, 23]])


def test_item_id_single(ml_ds: Dataset):
    items = ml_ds.items.index
    assert ml_ds.items.id(0) == items[0]
    assert ml_ds.items.id(ml_ds.item_count - 1) == items[-1]
    assert ml_ds.items.id(50) == items[50]


def test_item_id_many(ml_ds: Dataset):
    items = ml_ds.items.index
    assert np.all(ml_ds.items.ids([1, 5, 23]) == items[[1, 5, 23]])
    assert np.all(ml_ds.items.ids(np.array([1, 5, 23])) == items[[1, 5, 23]])


def test_user_num_single(ml_ds: Dataset):
    users = ml_ds.users.index
    assert ml_ds.users.number(users[0]) == 0
    assert ml_ds.users.number(users[50]) == 50


def test_user_num_many(ml_ds: Dataset):
    users = ml_ds.users.index
    assert np.all(ml_ds.users.numbers(users[[1, 5, 23]]) == [1, 5, 23])
    assert np.all(ml_ds.users.numbers(list(users[[1, 5, 23]])) == [1, 5, 23])


def test_user_num_missing_error(ml_ds: Dataset):
    with raises(KeyError):
        ml_ds.users.number(-402, missing="error")


def test_user_num_missing_negative(ml_ds: Dataset):
    assert ml_ds.users.number(-402, missing=None) is None


def test_user_num_missing_vector_negative(ml_ds: Dataset):
    u1 = ml_ds.users.index[5]
    u2 = ml_ds.users.index[100]
    res = ml_ds.users.numbers([u1, -402, u2], missing="negative")
    assert len(res) == 3
    assert np.all(res == [5, -1, 100])


def test_user_num_missing_vector_error(ml_ds: Dataset):
    u1 = ml_ds.users.index[5]
    u2 = ml_ds.users.index[100]
    with raises(KeyError):
        ml_ds.users.numbers([u1, -402, u2], missing="error")


def test_item_num_single(ml_ds: Dataset):
    items = ml_ds.items.index
    assert ml_ds.items.number(items[0]) == 0
    assert ml_ds.items.number(items[50]) == 50


def test_item_num_many(ml_ds: Dataset):
    items = ml_ds.items.index
    assert np.all(ml_ds.items.numbers(items[[1, 5, 23]]) == [1, 5, 23])
    assert np.all(ml_ds.items.numbers(list(items[[1, 5, 23]])) == [1, 5, 23])


def test_item_num_missing_error(ml_ds: Dataset):
    with raises(KeyError):
        ml_ds.items.number(-402, missing="error")



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import re
import time

from hypothesis import HealthCheck, given, settings
from hypothesis import strategies as st

from lenskit.logging import Stopwatch


def test_stopwatch_instant():
    w = Stopwatch()
    assert w.elapsed() > 0


def test_stopwatch_sleep():
    w = Stopwatch()
    time.sleep(0.5)
    assert w.elapsed() >= 0.45


def test_stopwatch_stop():
    w = Stopwatch()
    time.sleep(0.5)
    w.stop()
    time.sleep(0.5)
    assert w.elapsed() >= 0.45


def test_stopwatch_str():
    w = Stopwatch()
    time.sleep(0.5)
    s = str(w)
    assert s.endswith("ms")


def test_stopwatch_long_str():
    w = Stopwatch()
    time.sleep(1.2)
    s = str(w)
    assert s.endswith("s")


def test_stopwatch_minutes():
    w = Stopwatch()
    w.stop()
    assert w.stop_time is not None
    w.start_time = w.stop_time - 62
    s = str(w)
    p = re.compile(r"1m2.\d\ds")
    assert p.match(s)


def test_stopwatch_hours():
    w = Stopwatch()
    w.stop()
    assert w.stop_time is not None
    w.start_time = w.stop_time - 3663
    s = str(w)
    p = re.compile(r"1h1m3.\d\ds")
    assert p.match(s)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import click

from .summarize import summarize_profile


@click.group
def profile():
    """
    Summarize and manipulate profile data.
    """
    pass


profile.add_command(summarize_profile)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Rank discounting models.
"""

from abc import ABC, abstractmethod
from typing import Annotated

import numpy as np
from annotated_types import Gt, Lt
from pydantic import NonNegativeInt, PositiveFloat, validate_call

from lenskit.data.types import NPVector


class RankWeight(ABC):
    """
    Base class for rank weighting models.

    This returns *multiplicative* weights, such that scores should be multiplied
    by the weights in order to produce weighted scores.

    Stability:
        caller
    """

    @abstractmethod
    def weight(self, ranks: NPVector[np.int32]) -> NPVector[np.float64]:
        """
        Compute the discount for the specified ranks.

        Ranks must start with 1.
        """

    def log_weight(self, ranks: NPVector[np.int32]) -> NPVector[np.float64]:
        """
        Compute the (natural) log of the discount for the specified ranks.

        Ranks must start with 1.
        """
        return np.log(self.weight(ranks))

    def series_sum(self) -> float | None:
        """
        Get the sum of the infinite series of this discount function, if known.
        Some metrics (e.g. :func:`~lenskit.metrics.RBP`) will use this to
        normalize their measurements.
        """
        return None


class GeometricRankWeight(RankWeight):
    r"""
    Geometric cascade weighting for result ranks.

    This is the ranking model used by RBP :cite:p:`rbp`.

    For patience :math:`p`, the discount is given by :math:`p^{k-1}`.  The sum
    of this infinite series is :math:`\frac{1}{1 - p}`.

    Args:
        patience:
            The patience parameter :math:`p`.
    Stability:
        Caller
    """

    patience: float

    @validate_call
    def __init__(self, patience: Annotated[float, Gt(0.0), Lt(1.0)] = 0.85):
        self.patience = patience

    def weight(self, ranks) -> NPVector[np.float64]:
        return np.exp(self.log_weight(ranks))

    def log_weight(self, ranks) -> NPVector[np.float64]:
        return np.log(self.patience) * (ranks - 1)

    def series_sum(self) -> float:
        return 1 / (1 - self.patience)


class LogRankWeight(RankWeight):
    r"""
    Logarithmic weighting for result ranks, as used in NDCG.

    This is the ranking model typically used for DCG and NDCG.

    Since :math:`\operatorname{lg} 1 = 0`, simply taking the log will result in
    division by 0 when weights are applied.  The correction for this in the
    original NDCG paper :cite:p:`ndcg` is to clip the ranks, so that both of the
    first two positions have discount :math:`\operatorname{lg} 2`.  A different
    correction somtimes seen is to compute :math:`\operatorname{lg} (k+1)`. This
    discount supports both; the default is to clip, but if the ``offset`` option
    is set to a positive number, it is added to the ranks instead.

    Args:
        base:
            The log base to use.
        offset:
            An offset to add to ranks before computing logs.
    """

    base: float
    offset: int

    @validate_call
    def __init__(self, *, base: PositiveFloat = 2, offset: NonNegativeInt = 0):
        self.base = base
        self.offset = offset

    def weight(self, ranks):
        if self.offset > 0:
            return np.log(self.base) / np.log(ranks + self.offset)
        else:
            return np.log(self.base) / np.log(np.maximum(ranks, 2))



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import os
import sys
import warnings
from importlib.metadata import entry_points
from pathlib import Path

import click
import numpy as np

from lenskit import __version__, configure
from lenskit.config import locate_configuration_root
from lenskit.logging import LoggingConfig, console, get_logger

__all__ = ["lenskit", "main", "version"]
_log = get_logger(__name__)


def main():
    """
    Run the main LensKit CLI.  This just delegates to :fun:`lenskit`, but pretty-prints errors.
    """
    np.set_printoptions(threshold=20)
    try:
        ec = lenskit.main(standalone_mode=False)
    except click.ClickException as e:
        _log.error("CLI error, terminating: %s", e)
        sys.exit(2)
    except Exception as e:
        _log.error("LensKit command failed", exc_info=e)
        sys.exit(3)

    if isinstance(ec, int):
        sys.exit(ec)


@click.group("lenskit")
@click.option("-v", "--verbose", "verbosity", count=True, help="Enable verbose logging output")
@click.option(
    "-R",
    "--project-root",
    type=Path,
    metavar="DIR",
    help="Look for project root in DIR for configuration files.",
)
def lenskit(verbosity: int, project_root: Path | None):
    """
    Data and pipeline operations with LensKit.
    """

    # this code is run before any other command logic, so we can do global setup
    lc = LoggingConfig()
    if verbosity:
        lc.set_verbose(verbosity)
    lc.apply()

    if project_root is None:
        if pr := os.environ.get("LK_PROJECT_ROOT"):
            project_root = Path(pr)
        else:
            project_root = locate_configuration_root()

    configure(cfg_dir=project_root)


@lenskit.command("version")
def version():
    """
    Print LensKit version info.
    """
    console.print(f"LensKit version [bold cyan]{__version__}[/bold cyan].")


cli_plugins = entry_points(group="lenskit.cli.plugins")
for plugin in cli_plugins:
    try:
        cmd = plugin.load()
    except ImportError as e:
        warnings.warn(f"cannot load CLI {plugin.name}: {e}", ImportWarning)

    lenskit.add_command(cmd)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Item list collections.
"""

from ._base import ItemListCollection, ItemListCollector, MutableItemListCollection
from ._keys import GenericKey, UserIDKey, key_dict
from ._list import ListILC

__all__ = [
    "GenericKey",
    "UserIDKey",
    "ItemListCollection",
    "ItemListCollector",
    "MutableItemListCollection",
    "ListILC",
    "key_dict",
]



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from __future__ import annotations

from pathlib import Path

from pydantic import BaseModel, Field
from typing_extensions import Annotated, Literal, TypeAliasType

from lenskit.config import load_config_data
from lenskit.pipeline.config import PipelineConfig

SearchSpace = TypeAliasType("SearchSpace", "dict[str, SearchParam | SearchSpace]")


class SearchConfig(BaseModel):
    """
    Configuration options for the hyperparameter search.
    """

    method: Literal["optuna", "hyperopt", "random"] | None = None
    """
    The search method to use.
    """
    max_points: int = 60
    """
    The maximum number of points to try.
    """
    max_epochs: int = 100
    """
    The maximum number of epochs to use in iterative training.
    """
    min_epochs: int = 3
    """
    The minimum number of epochs for iterative training.
    """
    metric: str | None = None
    """
    The metric to use.
    """
    list_length: int | None = None
    """
    The length of recommendation lists to use.
    """
    num_cpus: int | Literal["threads", "all-threads"] = "threads"
    """
    The number of CPUs to request from Ray Tune.
    """
    num_gpus: int | float = 0
    """
    The number of GPUs to requrest from Ray Tune.
    """
    checkpoint_iters: int = 2
    """
    The frqeuencey for saving checkpoints.
    """


class TuningSpec(BaseModel, extra="forbid"):
    """
    Data model for hyperparameter tuning specifications.
    """

    @classmethod
    def load(cls, path: Path) -> TuningSpec:
        cfg = load_config_data(path, cls)
        cfg.file_path = path
        return cfg

    file_path: Annotated[Path | None, Field(exclude=True)] = None
    """
    The path to the spec file.
    """

    search: SearchConfig = SearchConfig()
    """
    Options for the hyperparameter search.
    """
    space: dict[str, SearchSpace]
    """
    The search space for tuning.
    """

    pipeline: PipelineFile | PipelineConfig
    """
    The pipeline to tune.
    """

    @property
    def component_name(self) -> str | None:
        """
        Get the name of the tuned component, if the search specifies
        parameters for a single component.
        """
        if len(self.space) == 1:
            for name in self.space.keys():
                return name
        else:
            return None

    def resolve_path(self, path: Path | str) -> Path:
        """
        Resolve a path relative to this specification's file.
        """
        if self.file_path is None:
            return Path(path)
        else:
            return self.file_path.parent / path


class PipelineFile(BaseModel, extra="forbid"):
    """
    Load a pipeline from a file.
    """

    file: Path
    """
    The file from whic to load the pipeline.
    """


class SearchParam(BaseModel):
    type: Literal["int", "float"]
    """
    The type of this parameter.
    """
    min: int | float
    """
    Minimum parameter value.
    """
    max: int | float
    """
    Maximum parameter value.
    """
    scale: Literal["uniform", "log"] = "uniform"
    """
    Search scale for parameter values.
    """
    base: float | None = None
    """
    Base for logarithmic search scales.
    """



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Data containers, the internal storage of data sets.
"""

from dataclasses import dataclass
from os import PathLike
from pathlib import Path
from shutil import rmtree

import pyarrow as pa
from pyarrow.parquet import read_table, write_table

from lenskit._accel import data as _data_accel
from lenskit.logging import get_logger
from lenskit.logging.stopwatch import Stopwatch

from .schema import DataSchema

_log = get_logger(__name__)


@dataclass(eq=False, order=False)
class DataContainer:
    """
    A general container for the data backing a dataset.
    """

    schema: DataSchema
    tables: dict[str, pa.Table]
    _sorted: bool = False
    _rel_coords: dict[str, _data_accel.CoordinateTable | None] | None = None

    def normalize(self):
        """
        Normalize data to adhere to the expectations of the most current schema.
        """

        self._sort_matrix_relationships()

    def _sort_matrix_relationships(self):
        log = _log.bind(dataset=self.schema.name)
        if self._sorted or self.schema.version >= "2025.3":
            return

        for name, rel in self.schema.relationships.items():
            if len(rel.entities) == 2 and not rel.repeats.is_present:
                # make sure they are a sorted matrix
                tbl = self.tables[name]
                e_cols = [e + "_num" for e in rel.entities.keys()]
                if not _data_accel.is_sorted_coo(tbl.to_batches(), *e_cols):
                    log.debug("sorting non-repeating relationship %s", name)
                    self.tables[name] = tbl.sort_by([(c, "ascending") for c in e_cols])

        self._sorted = True

    def __getstate__(self):
        # work around Ray's broken serialization of extension types
        state = dict(self.__dict__)
        state["tables"] = {name: _make_compat_table(tbl) for name, tbl in self.tables.items()}
        state["_rel_coords"] = None
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.tables = {name: _resolve_compat_table(tbl) for name, tbl in state["tables"].items()}

    def save(self, path: str | PathLike[str]):
        """
        Save the data to disk.
        """
        from .summary import save_stats

        path = Path(path)
        log = _log.bind(name=self.schema.name, path=str(path))
        log.info("saving dataset")

        if path.exists():
            log.warn("path already exists, removing")
            rmtree(path)

        log.debug("ensuring path exists")
        path.mkdir(exist_ok=True, parents=True)

        log.debug("writing schema")
        with open(path / "schema.json", "wt") as jsf:
            print(self.schema.model_dump_json(), file=jsf)

        for name, table in self.tables.items():
            log.debug("writing table", table=name, rows=table.num_rows)
            write_table(table, path / f"{name}.parquet", compression="zstd")

        log.debug("writing summary file")
        save_stats(self, path / "summary.md")

    @classmethod
    def load(cls, path: str | PathLike[str]):
        """
        Load data from disk.
        """
        path = Path(path)
        log = _log.bind(path=str(path))

        timer = Stopwatch()
        log.info("loading dataset")
        log.debug("reading schema")
        schema_file = path / "schema.json"
        schema = DataSchema.model_validate_json(schema_file.read_text(encoding="utf8"))
        log = log.bind(name=schema.name)

        tables = {}
        for name in schema.entities:
            log.debug("reading entity table %s", name, table=name, time=timer.elapsed())
            tables[name] = read_table(path / f"{name}.parquet").combine_chunks()

        for name in schema.relationships:
            log.debug("reading relationship table %s", name, table=name, time=timer.elapsed())
            tables[name] = read_table(path / f"{name}.parquet").combine_chunks()

        log.debug("finished loading data set in %s", timer, time=timer.elapsed())
        return cls(schema, tables)


def _make_compat_table(tbl: pa.Table):
    schema = tbl.schema
    ftypes = {}

    for name in schema.names:
        field = schema.field(name)
        if isinstance(field.type, pa.BaseExtensionType):
            ftypes[name] = field.type.storage_type
        else:
            ftypes[name] = field.type

    compat_schema = pa.schema(ftypes)
    return schema, tbl.cast(compat_schema, safe=True)


def _resolve_compat_table(data: tuple[pa.Schema, pa.Table]):
    schema, table = data
    return table.cast(schema)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import numpy as np
import pandas as pd

from lenskit.data.adapt import from_interactions_df


def test_item_subset(rng: np.random.Generator, ml_ratings: pd.DataFrame):
    items = rng.choice(ml_ratings["item_id"].unique(), 500, replace=False)
    ds = from_interactions_df(ml_ratings, items=items)

    assert ds.interaction_count < len(ml_ratings)
    assert ds.item_count == len(items)
    assert np.all(ds.items.ids() == np.sort(items))

    log = ds.interaction_table(format="pandas", original_ids=True)
    assert np.all(log["item_id"].isin(items))


def test_user_subset(rng: np.random.Generator, ml_ratings: pd.DataFrame):
    users = rng.choice(ml_ratings["user_id"].unique(), 500, replace=False)
    ds = from_interactions_df(ml_ratings, users=users)

    assert ds.interaction_count < len(ml_ratings)
    assert ds.user_count == len(users)
    assert np.all(ds.users.ids() == np.sort(users))

    log = ds.interaction_table(format="pandas", original_ids=True)
    assert np.all(log["user_id"].isin(users))



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Matrix layouts.
"""

# pyright: basic
from __future__ import annotations

import json
import warnings
from typing import Any, NamedTuple, TypeVar, cast

import numpy as np
import pyarrow as pa
import pyarrow.compute as pc
import scipy.sparse as sps
import torch
from numpy.typing import ArrayLike

from lenskit._accel import data as _data_accel

t = torch
M = TypeVar("M", "CSRStructure", sps.csr_array, sps.coo_array, sps.spmatrix, t.Tensor)

SPARSE_IDX_EXT_NAME = "lenskit.sparse_index"
SPARSE_IDX_LIST_EXT_NAME = "lenskit.sparse_index_list"
SPARSE_ROW_EXT_NAME = "lenskit.sparse_row"


class CSRStructure(NamedTuple):
    """
    Representation of the compressed sparse row structure of a sparse matrix,
    without any data values.

    Stability:
        Caller
    """

    rowptrs: np.ndarray
    colinds: np.ndarray
    shape: tuple[int, int]

    @property
    def nrows(self):
        return self.shape[0]

    @property
    def ncols(self):
        return self.shape[1]

    @property
    def nnz(self):
        return self.rowptrs[self.nrows]

    def extent(self, row: int) -> tuple[int, int]:
        return self.rowptrs[row], self.rowptrs[row + 1]

    def row_cs(self, row: int) -> np.ndarray:
        sp, ep = self.extent(row)
        return self.colinds[sp:ep]


class COOStructure(NamedTuple):
    """
    Representation of the coordinate structure of a sparse matrix, without any
    data values.

    Stability:
        Caller
    """

    row_numbers: np.ndarray
    col_numbers: np.ndarray
    shape: tuple[int, int]

    @property
    def nrows(self):
        return self.shape[0]

    @property
    def ncols(self):
        return self.shape[1]

    @property
    def nnz(self):
        return self.row_numbers.shape[0]


class SparseIndexType(pa.ExtensionType):
    """
    Data type for the index field of a sparse row.  Indexes are just stored as
    ``int32``s; the extension type attaches the row's dimensionality to the index
    field (making it easier to pass it to/from Rust, since we often pass arrays
    and not entire fields).

    .. stability:: internal
    """

    dimension: int

    def __init__(self, dimension: int):
        self.dimension = dimension
        super().__init__(pa.int32(), SPARSE_IDX_EXT_NAME)

    def check_dimension(self, expected: int | None) -> int:
        """
        Check that this index type has the expected dimension.

        Returns:
            The dimension of the index type.

        Raises:
            ValueError:
                If the type's dimension does not match the expected dimension.
        """
        if expected is not None and expected != self.dimension:
            raise ValueError(f"dimension mismatch: expected {expected}, found {self.dimension}")
        return self.dimension

    def __arrow_ext_serialize__(self) -> bytes:
        return json.dumps({"dimension": self.dimension}).encode()

    @classmethod
    def __arrow_ext_deserialize__(cls, storage_type, serialized):
        data = json.loads(serialized.decode())
        if not pa.types.is_int32(storage_type):
            raise TypeError("sparse index must be int32")
        return cls(data["dimension"])


class SparseIndexListType(pa.ExtensionType):
    """
    Sparse index lists.  These are the row type for structure-only sparse
    matrices.
    """

    value_type: None = None
    index_type: SparseIndexType

    def __init__(self, dimension: int, large: bool = False):
        self.index_type = SparseIndexType(dimension)

        lt_ctor = pa.large_list if large else pa.list_

        super().__init__(
            lt_ctor(self.index_type),
            SPARSE_IDX_LIST_EXT_NAME,
        )

    @classmethod
    def from_type(cls, data_type: pa.DataType, dimension: int | None = None) -> SparseIndexListType:
        """
        Create a sparse index list type from an Arrow data type, handling legacy struct
        layouts without the extension types.

        Args:
            data_type:
                The Arrow data type to interpret as a row type.
            dimension:
                The row dimension, if known from an external source.  If
                provided and the data type also includes the dimensionality,
                both dimensions must match.

        Raises:
            TypeError:
                If the data type is not a valid sparse row type.
            ValueError:
                If there is another error, such as mismatched dimensions.
        """
        if isinstance(data_type, SparseIndexListType):
            data_type.index_type.check_dimension(dimension)
            return data_type

        if pa.types.is_list(data_type):
            large = False
        elif pa.types.is_large_list(data_type):
            large = True
        else:
            raise TypeError(f"expected list type, found {data_type}")
        inner = data_type.value_type  # type: ignore

        dimension = _check_index_type(inner, dimension)

        return cls(dimension, large=large)

    @property
    def dimension(self) -> int:
        return self.index_type.dimension

    def __arrow_ext_serialize__(self) -> bytes:
        return b""

    @classmethod
    def __arrow_ext_deserialize__(cls, storage_type, serialized):
        return cls.from_type(storage_type)

    def __arrow_ext_class__(self):
        return SparseRowArray


class SparseRowType(pa.ExtensionType):
    """
    Data type for sparse rows stored in Arrow.  Sparse rows are stored as lists
    of structs with ``index`` and ``column`` fields.

    .. stability:: internal
    """

    value_type: pa.DataType | None
    index_type: SparseIndexType

    def __init__(
        self, dimension: int, value_type: pa.DataType | None = pa.float32(), large: bool = False
    ):
        self.value_type = value_type
        self.index_type = SparseIndexType(dimension)

        lt_ctor = pa.large_list if large else pa.list_

        if value_type is None:
            element_type = self.index_type
        else:
            element_type = pa.struct(
                [
                    ("index", self.index_type),
                    ("value", value_type),
                ]
            )

        super().__init__(
            lt_ctor(element_type),
            SPARSE_ROW_EXT_NAME,
        )

    @classmethod
    def from_type(cls, data_type: pa.DataType, dimension: int | None = None) -> SparseRowType:
        """
        Create a sparse row type from an Arrow data type, handling legacy struct
        layouts without the extension types.

        Args:
            data_type:
                The Arrow data type to interpret as a row type.
            dimension:
                The row dimension, if known from an external source.  If
                provided and the data type also includes the dimensionality,
                both dimensions must match.

        Raises:
            TypeError:
                If the data type is not a valid sparse row type.
            ValueError:
                If there is another error, such as mismatched dimensions.
        """
        if isinstance(data_type, SparseRowType):
            data_type.index_type.check_dimension(dimension)
            return data_type

        if pa.types.is_list(data_type):
            large = False
        elif pa.types.is_large_list(data_type):
            large = True
        else:
            raise TypeError(f"expected list type, found {data_type}")
        inner = data_type.value_type  # type: ignore

        if not pa.types.is_struct(inner):
            raise TypeError(f"expected struct type, found {inner}")

        if inner.num_fields != 2:
            raise TypeError(f"element struct must have 2 elements, found {inner.num_fields}")

        idx_f = inner.field(0)
        assert isinstance(idx_f, pa.Field)
        if idx_f.name != "index":
            raise TypeError(f"first field of element struct must be 'index', found {idx_f.name}")

        dimension = _check_index_type(idx_f.type, dimension)

        val_f = inner.field(1)
        if val_f.name != "value":
            raise TypeError(f"second field of element struct must be 'value', found {val_f.name}")

        return cls(dimension, val_f.type, large=large)

    @property
    def dimension(self) -> int:
        return self.index_type.dimension

    def __arrow_ext_serialize__(self) -> bytes:
        return b""

    @classmethod
    def __arrow_ext_deserialize__(cls, storage_type, serialized):
        return cls.from_type(storage_type)

    def __arrow_ext_class__(self):
        return SparseRowArray


class SparseRowArray(pa.ExtensionArray):
    """
    An array of sparse rows (a compressed sparse row matrix).

    .. stability:: internal
    """

    type: SparseRowType | SparseIndexListType

    @classmethod
    def from_arrays(
        cls,
        offsets: ArrayLike,
        indices: ArrayLike,
        values: ArrayLike | None = None,
        *,
        shape: tuple[int, int] | None = None,
    ) -> SparseRowArray:
        offsets = pa.array(offsets)
        large = pa.types.is_int64(offsets.type)
        indices = pa.array(indices, type=pa.int32())

        if shape:
            _nr, nc = shape
        else:
            nc = pc.max(indices).to_py() + 1

        if values is not None:
            values = pa.array(values)
            row_type = SparseRowType(nc, values.type, large)
            index_type = row_type.index_type

            elements = pa.StructArray.from_arrays(
                [indices.cast(index_type), values], names=["index", "value"]
            )
        else:
            row_type = SparseIndexListType(nc, large)
            elements = indices.cast(row_type.index_type)

        if large:
            storage = pa.LargeListArray.from_arrays(offsets, elements)
        else:
            storage = pa.ListArray.from_arrays(offsets, elements)

        return pa.ExtensionArray.from_storage(row_type, storage)  # type: ignore

    @classmethod
    def from_array(cls, array: pa.Array, dimension: int | None = None) -> SparseRowArray:
        """
        Interpret an Arrow array as a sparse row array, if possible.  Handles
        legacy layouts without the extension types.

        Args:
            array:
                The array to convert.
            dimension:
                The dimensionality of the sparse rows, if known from an external source.
        """
        if isinstance(array, SparseRowArray):
            assert isinstance(array.type, SparseRowType)
            array.type.index_type.check_dimension(dimension)
            return array
        elif isinstance(array.type, SparseRowType):
            array.type.index_type.check_dimension(dimension)
            return pa.ExtensionArray.from_storage(array.type, array)  # type: ignore

        et = SparseRowType.from_type(array.type, dimension)
        return pa.ExtensionArray.from_storage(et, array.cast(et.storage_type))  # type: ignore

    @classmethod
    def from_scipy(
        cls,
        matrix: sps.sparray,
        *,
        values: bool = True,
        large: bool | None = None,
    ) -> SparseRowArray:
        """
        Create a sparse row array from a SciPy sparse matrix.

        Args:
            csr:
                The SciPy sparse matrix (in CSR format).
            values:
                Whether to include the values or create a structure-only array.
            large:
                ``True`` to force creation of a :class:`pa.LargeListArray`.

        Returns:
            The sparse row array.
        """
        matrix = cast("sps.csr_array[Any, tuple[int, int]]", matrix.tocsr())  # type: ignore
        _nr, dim = matrix.shape
        smax = np.iinfo(np.int32).max

        offsets = matrix.indptr
        if large:
            offsets = np.require(offsets, np.int64)
        elif matrix.nnz < smax:
            offsets = np.require(offsets, dtype=np.int32)
        elif large is False:
            raise ValueError("sparse matrix size {:,d} too large for list".format(matrix.nnz))
        cols = pa.array(matrix.indices, SparseIndexType(dim))

        vals = pa.array(matrix.data) if values else None

        return cls.from_arrays(offsets, cols, vals, shape=matrix.shape)

    def to_scipy(self) -> sps.csr_array[Any, tuple[int, int]]:
        """
        Convert this sparse row array to a SciPy sparse array.
        """
        if not self.has_values:
            raise TypeError("structure-only arrays cannot convert to scipy")
        return sps.csr_array(
            (self.values.to_numpy(), self.indices.to_numpy(), self.offsets.to_numpy()),
            shape=(len(self), self.type.dimension),
        )

    def to_torch(self) -> torch.Tensor:
        """
        Convert this sparse row array to a Torch sparse tensor.
        """
        if not self.has_values:
            raise TypeError("structure-only arrays cannot convert to Torch")
        return torch.sparse_csr_tensor(
            crow_indices=torch.as_tensor(
                self.offsets.to_numpy(zero_copy_only=False, writable=True)
            ),
            col_indices=torch.as_tensor(self.indices.to_numpy(zero_copy_only=False, writable=True)),
            values=torch.as_tensor(self.values.to_numpy(zero_copy_only=False, writable=True)),
            size=self.shape,
        )

    def to_coo(self) -> pa.Table:
        """
        Convert this array to table representing the array in COO format.
        """
        nr, nc = self.shape
        lengths = self.storage.value_lengths()
        assert len(lengths) == nr
        rowinds = np.repeat(np.arange(nr, dtype=np.int32), lengths)
        fields = {"row": rowinds, "col": self.indices}
        if self.has_values:
            fields["value"] = self.values
        return pa.table(fields)

    @property
    def dimension(self) -> int:
        """
        Get the number of columns in the sparse matrix.
        """
        return self.type.dimension

    @property
    def shape(self) -> tuple[int, int]:
        return (len(self), self.dimension)

    @property
    def has_values(self) -> bool:
        return self.type.value_type is not None

    @property
    def offsets(self) -> pa.Int32Array:
        return self.storage.offsets

    @property
    def indices(self) -> pa.Int32Array:
        if self.has_values:
            return self.storage.values.field(0)
        else:
            return self.storage.values

    @property
    def values(self) -> pa.Array | None:
        if self.has_values:
            return self.storage.values.field(1)
        else:
            return None

    def structure(self) -> SparseRowArray:
        """
        Get the structure of this matrix (without values).
        """
        if self.has_values:
            return self.from_arrays(self.offsets, self.indices, shape=self.shape)
        else:
            return self

    def transpose(self) -> SparseRowArray:
        """
        Get the transpose of this sparse matrix.
        """
        nr, nc = self.shape
        row_ptr, col_ind, perm = _data_accel.transpose_csr(self.structure(), self.has_values)
        if perm is None:
            return self.from_arrays(row_ptr, col_ind, shape=(nc, nr))
        else:
            values = self.values
            assert values is not None
            values = values.take(perm)
            return self.from_arrays(row_ptr, col_ind, values, shape=(nc, nr))

    def row_extent(self, row: int) -> tuple[int, int]:
        """
        Get the start and end of a row.
        """
        start = self.storage.offsets[row].as_py()
        end = self.storage.offsets[row + 1].as_py()
        return start, end

    def row_indices(self, row: int) -> pa.Int32Array:
        """
        Get the index array for a compressed sparse row.
        """
        sp, ep = self.row_extent(row)
        indices = self.indices
        return indices.slice(sp, ep - sp)


def _check_index_type(index_type: pa.DataType, dimension: int | None) -> int:
    if isinstance(index_type, SparseIndexType):
        return index_type.check_dimension(dimension)
    elif dimension is None:
        raise TypeError(
            f"index type must be SparseIndex when no external dimension, found {index_type}"
        )
    elif not pa.types.is_int32(index_type):
        if pa.types.is_int64(index_type):
            warnings.warn("sparse row has legacy int64 indices", DeprecationWarning)
        else:
            raise TypeError(f"index type must be SparseIndex or int32, found {index_type}")

    return dimension


pa.register_extension_type(SparseIndexType(0))  # type: ignore
pa.register_extension_type(SparseIndexListType(0))  # type: ignore
pa.register_extension_type(SparseRowType(0))  # type: ignore



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import logging

import numpy as np
import pandas as pd

import hypothesis.strategies as st
from hypothesis import given
from pytest import approx, mark

from lenskit.data import ItemList
from lenskit.metrics import call_metric
from lenskit.metrics.ranking import RBP, LogRankWeight
from lenskit.testing import demo_recs, integer_ids  # noqa: F401

_log = logging.getLogger(__name__)


def test_rbp_empty():
    recs = ItemList([], ordered=True)
    truth = ItemList([1, 2, 3])
    assert call_metric(RBP, recs, truth) == approx(0.0)


def test_rbp_no_match():
    recs = ItemList([4], ordered=True)
    truth = ItemList([1, 2, 3])
    assert call_metric(RBP, recs, truth) == approx(0.0)


def test_rbp_one_match():
    recs = ItemList([1], ordered=True)
    truth = ItemList([1, 2, 3])
    assert call_metric(RBP, recs, truth) == approx(0.15)


@given(st.lists(integer_ids(), min_size=1, max_size=100, unique=True), st.floats(0.05, 0.95))
def test_rbp_perfect(items, p):
    n = len(items)
    recs = ItemList(items, ordered=True)
    truth = ItemList(items)
    assert call_metric(RBP, recs, truth, patience=p) == approx(np.sum(p ** np.arange(n)) * (1 - p))


@given(st.lists(integer_ids(), min_size=1, max_size=100, unique=True), st.floats(0.05, 0.95))
def test_rbp_perfect_norm(items, p):
    recs = ItemList(items, ordered=True)
    truth = ItemList(items)
    assert call_metric(RBP, recs, truth, patience=p, normalize=True) == approx(1.0)


@given(
    st.lists(integer_ids(), min_size=1, max_size=100, unique=True),
    st.integers(1, 100),
    st.floats(0.05, 0.95),
)
def test_rbp_perfect_k(items, k, p):
    n = len(items)
    eff_n = min(n, k)
    recs = ItemList(items, ordered=True)
    truth = ItemList(items)
    assert call_metric(RBP, recs, truth, k, patience=p) == approx(
        np.sum(p ** np.arange(eff_n)) * (1 - p)
    )


@given(
    st.lists(integer_ids(), min_size=1, max_size=100, unique=True),
    st.integers(1, 100),
)
def test_rbp_perfect_log_weight(items, k):
    recs = ItemList(items, ordered=True)
    truth = ItemList(items)
    assert call_metric(RBP, recs, truth, k, weight=LogRankWeight(offset=1)) == approx(1.0)


@given(
    st.lists(integer_ids(), min_size=2, max_size=100, unique=True),
)
def test_rbp_partial_log_weight(items):
    recs = ItemList(items, ordered=True)
    truth = ItemList(items[::2])
    w = np.reciprocal(np.log2(np.arange(1, len(recs) + 1) + 1))
    gw = np.zeros_like(w)
    gw[::2] = w[::2]
    assert call_metric(RBP, recs, truth, weight=LogRankWeight(offset=1)) == approx(
        np.sum(gw) / np.sum(w)
    )


@given(
    st.lists(integer_ids(), min_size=1, max_size=100, unique=True),
    st.integers(1, 100),
    st.floats(0.05, 0.95),
)
def test_rbp_perfect_k_norm(items, k, p):
    recs = ItemList(items, ordered=True)
    truth = ItemList(items)
    assert call_metric(RBP, recs, truth, k, patience=p, normalize=True) == approx(1.0)


def test_rbp_missing():
    recs = ItemList([1, 2], ordered=True)
    truth = ItemList([1, 2, 3])
    # (1 + 0.5) * 0.5
    assert call_metric(RBP(patience=0.5), recs, truth) == approx(0.75)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import pickle
from pathlib import Path

import click
from xopen import xopen

from lenskit.data import Dataset
from lenskit.logging import get_logger
from lenskit.pipeline import Component, Pipeline, topn_pipeline
from lenskit.pipeline.types import resolve_type_string

_log = get_logger(__name__)


@click.command("train")
@click.option("-C", "--scorer-class", metavar="CLS", help="Train a model with scorer CLS.")
@click.option(
    "-c", "--config", type=Path, metavar="FILE", help="Load pipeline configuration from FILE."
)
@click.option(
    "-o",
    "--output",
    "out_file",
    metavar="FILE",
    default="model.pkl",
    help="Output file for trained model.",
)
@click.option("--name", help="Name of the recommendation pipeline.")
@click.option(
    "--rating-predictor",
    is_flag=True,
    help="Include rating prediction in the pipeline capabilities.",
)
@click.option("-n", "--list-length", type=int, help="Default list length for pipeline ranker.")
@click.argument("dataset", metavar="DATA", type=Path)
def train(
    scorer_class: str | None,
    config: Path | None,
    out_file: Path,
    name: str | None,
    list_length: int | None,
    rating_predictor: bool,
    dataset: Path,
):
    """
    Train a recommendation pipeline and serialize it to disk.
    """
    _log.warning("the training CLI is experimental and may change without notice")

    if config is not None:
        if scorer_class is not None:
            _log.error("cannot specify both scorer class and configuration file")
            raise SystemExit(3)

        pipe = Pipeline.load_config(config)

    elif scorer_class is not None:
        scorer = resolve_type_string(scorer_class)
        if name is None:
            name = scorer.__name__
        assert issubclass(scorer, Component)
        pipe = topn_pipeline(scorer, predicts_ratings=rating_predictor, n=list_length, name=name)

    else:
        _log.error("no scorer specified")
        raise SystemExit(5)

    data = Dataset.load(dataset)
    log = _log.bind(data=data.name, name=name)

    log.info("training model")
    pipe.train(data)

    _log.info("saving trained model", file=out_file)
    with xopen(out_file, "wb", threads=0) as pf:
        pickle.dump(pipe, pf)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from __future__ import annotations

import math
from collections.abc import Callable
from typing import Literal, TypeAlias

import numpy as np
import torch
from pydantic import PositiveFloat, PositiveInt, model_validator
from torch.nn import functional as F

from lenskit.data import Dataset

from ._base import FlexMFConfigBase, FlexMFScorerBase
from ._model import FlexMFModel
from ._training import FlexMFTrainerBase, FlexMFTrainingBatch, FlexMFTrainingData

MAX_TRIES = 200
ImplicitLoss: TypeAlias = Literal["logistic", "pairwise", "warp"]
NegativeStrategy: TypeAlias = Literal["uniform", "popular", "misranked"]


class FlexMFImplicitConfig(FlexMFConfigBase):
    """
    Configuration for :class:`FlexMFImplicitScorer`.  It inherits base model
    options from :class:`FlexMFConfigBase`.

    Stability:
        Experimental
    """

    loss: ImplicitLoss = "logistic"
    """
    The loss to use for model training.
    """

    negative_strategy: NegativeStrategy | None = None
    """
    The negative sampling strategy.  The default is ``"misranked"`` for WARP
    loss and ``"uniform"`` for other losses.
    """

    negative_count: PositiveInt = 1
    """
    The number of negative items to sample for each positive item in the
    training data.  With BPR loss, the positive item is compared to each
    negative item; with logistic loss, the positive item is treated once per
    learning round, so this setting effectively makes the model learn on _n_
    negatives per positive, rather than giving positive and negative examples
    equal weight.
    """

    positive_weight: PositiveFloat = 1.0
    """
    A weighting multiplier to apply to the positive item's loss, to adjust the
    relative importance of positive and negative classifications.  Only applies
    to logistic loss.
    """

    user_bias: bool | None = None
    """
    Whether to learn a user bias term.  If unspecified, the default depends on
    the loss function (``False`` for pairwise and ``True`` for logistic).
    """
    item_bias: bool = True
    """
    Whether to learn an item bias term.
    """

    def selected_negative_strategy(self) -> NegativeStrategy:
        if self.negative_strategy is not None:
            return self.negative_strategy
        elif self.loss == "warp":
            return "misranked"
        else:
            return "uniform"

    @model_validator(mode="after")
    def check_strategies(self):
        if (
            self.loss == "warp"
            and self.negative_strategy is not None
            and self.negative_strategy != "misranked"
        ):
            raise ValueError("WARP loss requires “misranked” negative strategy")

        if self.selected_negative_strategy() == "misranked" and self.negative_count > 1:
            raise ValueError("misrank negatives only works with single negatives")

        return self


class FlexMFImplicitScorer(FlexMFScorerBase):
    """
    Implicit-feedback rating prediction with FlexMF.  This is capable of
    realizing multiple models, including:

    - BPR-MF (Bayesian personalized ranking) :cite:p:`BPR` (with ``"pairwise"`` loss)
    - Logistic matrix factorization :cite:p:`LogisticMF` (with ``"logistic"`` loss)

    All use configurable negative sampling, including the sampling approach from WARP.

    Stability:
        Experimental
    """

    config: FlexMFImplicitConfig

    def create_trainer(self, data, options):
        if self.config.selected_negative_strategy() == "misranked":
            return FlexMFWARPTrainer(self, data, options)
        else:
            return FlexMFImplicitTrainer(self, data, options)


class FlexMFImplicitTrainer(FlexMFTrainerBase[FlexMFImplicitScorer, FlexMFImplicitConfig]):
    _loss: Callable[[torch.Tensor, torch.Tensor, float, torch.Tensor | None], torch.Tensor]

    def __init__(self, component, data, options):
        super().__init__(component, data, options)
        match self.config.loss:
            case "logistic":
                self._loss = _loss_logistic
            case "pairwise":
                self._loss = _loss_pairwise
            case "warp":
                self._loss = _loss_warp
            case other:  # pragma: nocover
                raise ValueError(f"unknown loss {other}")

    def prepare_data(self, data: Dataset) -> FlexMFTrainingData:
        """
        Set up the training data and context for the scorer.
        """

        matrix = data.interactions().matrix()
        coo = matrix.coo_structure()

        # save data we learned at this stage
        self.component.users = data.users
        self.component.items = data.items

        return FlexMFTrainingData(
            batch_size=self.config.batch_size,
            n_users=data.user_count,
            n_items=data.item_count,
            users=coo.row_numbers,
            items=coo.col_numbers,
            matrix=matrix,
        )

    def create_model(self) -> FlexMFModel:
        """
        Prepare the model for training.
        """
        user_bias = self.config.user_bias
        if user_bias is None:
            if self.config.loss == "logistic":
                user_bias = True
            else:
                user_bias = False

        return FlexMFModel(
            self.config.embedding_size,
            self.data.n_users,
            self.data.n_items,
            self.torch_rng,
            user_bias=user_bias,
            item_bias=self.config.item_bias,
            sparse=self.config.reg_method != "AdamW",
        )

    def score(self, users, items) -> tuple[torch.Tensor, torch.Tensor]:
        if self.explicit_norm:
            result = self.model(users, items, return_norm=True)
            scores = result[0, ...]

            norms = result[1, ...]
        else:
            scores = self.call_model(users, items, return_norm=False)
            norms = torch.tensor(0.0)

        return scores, norms

    def train_batch(self, batch: FlexMFTrainingBatch) -> torch.Tensor:
        users = torch.as_tensor(batch.users.reshape(-1, 1)).to(self.device, non_blocking=True)
        positives = torch.as_tensor(batch.items.reshape(-1, 1)).to(self.device, non_blocking=True)

        pos_pred, pos_norm = self.score(users, positives)

        neg_items, neg_pred, neg_norm, weights = self.scored_negatives(batch, users, pos_pred)

        loss = self._loss(pos_pred, neg_pred, self.config.positive_weight, weights)

        if self.explicit_norm:
            loss = loss + self.config.regularization * 0.5 * (pos_norm.mean() + neg_norm.mean())

        loss.backward()
        self.opt.step()

        return loss

    def scored_negatives(
        self, batch: FlexMFTrainingBatch, users: torch.Tensor, pos_scores: torch.Tensor
    ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor | None]:
        assert batch.data.matrix is not None
        assert isinstance(batch.users, np.ndarray)
        items = torch.as_tensor(
            batch.data.matrix.sample_negatives(
                batch.users,
                weighting=self.config.selected_negative_strategy(),
                n=self.config.negative_count,
                rng=self.rng,
            )
        ).to(users.device, non_blocking=True)
        scores, norms = self.score(users, items)
        return items, scores, norms, None


class FlexMFWARPTrainer(FlexMFImplicitTrainer):
    def scored_negatives(self, batch, users, pos_scores):
        assert batch.data.matrix is not None
        assert isinstance(batch.users, np.ndarray)

        # start looking for misranked models
        idx_range = torch.arange(len(users), device=users.device)
        neg_scores = torch.full((len(users),), -math.inf, device=users.device)
        neg_norms = torch.zeros(len(users), device=users.device)
        neg_counts = torch.zeros(len(users))
        neg_items = torch.empty(len(users), dtype=torch.int32, device=users.device)
        needed = neg_counts <= 0
        tries = 0
        while torch.any(needed):
            bi = tries % 10
            tries += 1
            if tries > MAX_TRIES:
                self.log.debug("exceed MAX_TRIES for %d items", np.sum(needed.numpy()))
                continue

            n_dev = needed.to(users.device)
            n_users = batch.users[needed]
            if bi == 0:
                neg_cand = torch.as_tensor(
                    batch.data.matrix.sample_negatives(
                        n_users,
                        n=10,
                        weighting="uniform",
                        rng=self.rng,
                    )
                ).to(users.device)
                nc_scores, nc_norms = self.score(users[n_dev], neg_cand)

            found = nc_scores[:, bi] > pos_scores[n_dev, 0]
            f_idx = idx_range[n_dev][found]
            neg_items[f_idx] = neg_cand[found, bi]
            neg_scores[f_idx] = nc_scores[found, bi]
            if nc_norms.shape:
                neg_norms[f_idx] = nc_norms[found, bi]

            nf = ~found
            nf_idx = idx_range[n_dev][nf]
            nf_big = nc_scores[nf, bi] > neg_scores[nf_idx]
            nf_upd = nf_idx[nf_big]
            # assert nf_big.shape == neg_cand.shape[:1], f"{nf_big.shape} != {neg_cand.shape}"
            neg_items[nf_upd] = neg_cand[nf, bi][nf_big]
            neg_scores[nf_upd] = nc_scores[nf, bi][nf_big]
            if nc_norms.shape:
                neg_norms[nf_upd] = nc_norms[nf, bi][nf_big]

            neg_counts[needed] += 1

            needed = neg_counts <= 0

        ranks = ((self.data.n_items - 1) / (neg_counts + 1)).to(torch.float64).detach()
        # L(k) = sum i=1..k 1/i = harmonic k
        # approximate harmonic k with log
        weights = (
            torch.log(ranks)
            + np.euler_gamma
            + 1 / (2 * ranks)
            - 1 / (12 * ranks**2)
            + 1 / (120 * ranks**4)
        )
        return neg_items, neg_scores.reshape(-1, 1), neg_norms, weights.to(pos_scores.device)


def _loss_logistic(pos_pred, neg_pred, pos_weight, weights):
    pos_lp = -F.logsigmoid(pos_pred) * pos_weight
    neg_lp = -F.logsigmoid(-neg_pred)
    tot_lp = pos_lp.sum() + neg_lp.sum()
    tot_n = pos_lp.nelement() + neg_lp.nelement()
    return tot_lp / tot_n


def _loss_pairwise(pos_pred, neg_pred, pos_weight, weights):
    lp = -F.logsigmoid(pos_pred - neg_pred)
    return lp.mean()


def _loss_warp(pos_pred, neg_pred, pos_weight, weights):
    assert weights is not None
    lp = -F.logsigmoid(pos_pred - neg_pred) * weights
    return lp.mean()



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from pathlib import Path

from pytest import importorskip, mark, skip

from lenskit.data import Dataset, load_movielens
from lenskit.parallel.ray import ensure_cluster

ray = importorskip("ray")

data_dir = Path("data")


@mark.parametrize("name", ["ml-latest-small", "ml-100k.zip", "ml-20m.zip"])
def test_ray_roundtrip(name):
    ensure_cluster()
    ds_path = data_dir / name
    if not ds_path.exists():
        skip(f"dataset {name} not available")

    ds = load_movielens(ds_path)

    ref = ray.put(ds)
    ds2 = ray.get(ref)

    assert isinstance(ds2, Dataset)
    assert ds2 is not ds

    assert ds2.item_count == ds.item_count



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from __future__ import annotations

import warnings
from abc import ABC, abstractmethod
from collections.abc import Sequence
from os import PathLike
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Generator,
    Generic,
    Iterator,
    Literal,
    Mapping,
    Protocol,
    overload,
)

import pandas as pd
import pyarrow as pa
from more_itertools import chunked
from pyarrow.parquet import ParquetDataset, ParquetWriter

from lenskit.diagnostics import DataWarning

from ..arrow import explode_column
from ..builder import DatasetBuilder
from ..container import DataContainer
from ..items import ItemList
from ..types import ID, Column
from ._keys import KL, GenericKey, K, create_key_type, key_dict, key_fields, project_key

if TYPE_CHECKING:
    from ..dataset import Dataset


class ItemListCollection(Generic[KL], ABC):
    """
    A collection of item lists.  This protocol defines read access to the
    collection; see :class:`ItemListCollector` for the ability to add new lists.
    See :ref:`item-list-collections` for an introduction to using this class.

    An item list collection consists of a sequence of item lists with associated
    *keys* following a fixed schema.  Item list collections support iteration
    (in order) and lookup by key. They are used to represent a variety of
    things, including test data and the results of a batch run. The length of an
    item list collection (accessed with :func:`len`) is the number of item lists
    in the collection, not the total number of items in the lists.

    The key schema can be specified either by a list of field names, or by
    providing a named tuple class (created by either :func:`namedtuple` or
    :class:`NamedTuple`) defining the key schema.  Schemas should **not** be
    nested: field values must be scalars, not tuples or lists.  Keys should also
    be hashable.

    This protocol and its implementations exist, instead of using raw
    dictionaries or lists, to consistently handle some of the nuances of
    multi-valued keys, and different collections having different key fields.
    For example, if a run produces item lists with both user IDs and sequence
    numbers, but your test data is only indexed by user ID, the *projected
    lookup* capabilities make it easy to find the test data to go with an item
    list in the run.

    Item list collections support indexing by position, like a list, returning a
    tuple of the key and list; iterating over an item list collection similarly
    produces ``(key, list)`` pairs (so an item list collection is a
    :class:`~collections.abc.Sequence` of key/list pairs).

    If the item list is _indexed_ (constructed with ``index=True``), it also
    supports lookup by _key_ with :meth:`lookup`.  The key can be supplied as
    either a tuple or an instance of the key type.  If more than one item with
    the same key is inserted into the collection, then the _last_ one is
    returned (just like a dictionary), but the others remain in the underlying
    list when it is iterated.

    .. note::

        Constructing an item list collection yields a
        :class:`~lenskit.data.ListILC`.

    Args:
        key:
            The type (a NamedTuple class) or list of field names specifying the
            key schema.
        index:
            Whether or not to index lists by key to facilitate fast lookups.
    """

    _key_class: type[KL]

    def __new__(cls, key: type[KL] | Sequence[str] | None = None, *, index: bool = True):
        if key is not None and (cls == ItemListCollection or cls == MutableItemListCollection):
            return cls.empty(key, index=index)
        else:
            return super().__new__(cls)

    def __init__(self, key: type[KL] | Sequence[str]):
        if isinstance(key, type):
            self._key_class = key
        else:
            self._key_class = create_key_type(*key)  # type: ignore

    @staticmethod
    def empty(key: type[K] | Sequence[str], *, index: bool = True) -> MutableItemListCollection[K]:
        """
        Create a new empty, mutable item list collection.
        """
        from ._list import ListILC

        return ListILC(key, index=index)

    @overload
    @staticmethod
    def from_dict(
        data: Mapping[GenericKey | ID, ItemList], key: type[K]
    ) -> ItemListCollection[K]: ...
    @overload
    @staticmethod
    def from_dict(
        data: Mapping[GenericKey | ID, ItemList], key: Sequence[str] | str | None = None
    ) -> ItemListCollection[GenericKey]: ...
    @staticmethod
    def from_dict(
        data: Mapping[GenericKey | ID, ItemList],
        key: type[K] | Sequence[str] | str | None = None,
    ) -> ItemListCollection[GenericKey]:
        """
        Create an item list collection from a dictionary.

        .. seealso::
            :meth:`lenskit.data.collection.ListILC.from_dict`
        """
        from ._list import ListILC

        return ListILC.from_dict(data, key)

    @staticmethod
    def from_df(
        df: pd.DataFrame, key: type[K] | Sequence[Column] | Column | None = None, *others: Column
    ) -> MutableItemListCollection[Any]:
        """
        Create an item list collection from a data frame.

        .. seealso::
            :meth:`lenskit.data.collection.ListILC.from_df`

        .. note::

            Keys with empty item lists will be silently excluded from the output
            data.

        Args:
            df:
                The data frame to convert.
            key:
                The key type or field(s).  Can be specified as a single column
                name (or :class:`~lenskit.data.types.AliasedColumn`).
            others:
                Other columns to consider; primarily used to pass additional
                aliased columns to normalize other clumnes like the item ID.
        """
        from ._list import ListILC

        return ListILC.from_df(df, key)

    @staticmethod
    def from_arrow(table: pa.Table) -> MutableItemListCollection[Any]:
        """
        Convert an Arrow table into an item list collection.  The table must be
        in ``'native`'' format.
        """
        from ._list import ListILC

        return ListILC.from_arrow(table)

    def to_df(self) -> pd.DataFrame:
        """
        Convert this item list collection to a data frame.

        .. warning::

            If this item list collection has any keys with empty lists, those
            lists will be excluded from the output.
        """
        if any(len(il) == 0 for il in self.lists()):
            warnings.warn(
                "item list collection has empty lists, they will be dropped",
                DataWarning,
                stacklevel=2,
            )
        fields = list(self.key_fields)
        return (
            pd.concat({k: il.to_df(numbers=False) for (k, il) in self.items()}, names=fields)
            .reset_index(fields)
            .reset_index(drop=True)
        )

    def to_arrow(
        self, *, batch_size: int = 5000, layout: Literal["native", "flat"] = "native"
    ) -> pa.Table:
        """
        Convert this item list collection to an Arrow table.

        The resulting table has one row per item list, with the item list
        contents an ``items`` column of a structured list type.  This preserves
        empty item lists for higher-fidelity data storage.

        Args:
            batch_size:
                The Arrow record batch size.
        """
        return pa.Table.from_batches(self.record_batches(batch_size=batch_size, layout=layout))

    @overload
    def to_dataset(
        self,
        class_name: str = "interaction",
        *,
        result: Literal["dataset"] = "dataset",
    ) -> Dataset: ...
    @overload
    def to_dataset(
        self,
        class_name: str = "interaction",
        *,
        result: Literal["container"],
    ) -> DataContainer: ...
    @overload
    def to_dataset(
        self,
        class_name: str = "interaction",
        *,
        result: Literal["builder"],
    ) -> DatasetBuilder: ...
    def to_dataset(
        self,
        class_name: str = "interaction",
        *,
        result: Literal["dataset", "container", "builder"] = "dataset",
    ):
        """
        Construct a dataset populated with this item list collection's data as
        interactions.

        Args:
            cls:
                The interaction class name.
            result:
                Whether to return a fully-instantiated dataset, a container,
                or a dataset builder.
        """
        dsb = DatasetBuilder()
        entities = [k[:-3] for k in self.key_fields if k.endswith("_id")]
        data = self.to_arrow(layout="flat")

        for ecls in entities:
            dsb.add_entity_class(ecls)
            # speed up?
            dsb.add_entities(ecls, list(set(getattr(k, f"{ecls}_id") for k in self.keys())))

        entities += ["item"]
        dsb.add_interactions(class_name, data, entities=entities, missing="insert")

        match result:
            case "builder":
                return dsb
            case "container":
                return dsb.build_container()
            case "dataset":
                return dsb.build()
            case _:  # pragma: nocover
                raise ValueError(f"invalid result type {result}")

    def save_parquet(
        self,
        path: PathLike[str],
        *,
        layout: Literal["native", "flat"] = "native",
        batch_size: int = 5000,
        compression: Literal["zstd", "gzip", "snappy", "lz4"] | None = "zstd",
        mkdir: bool = True,
    ) -> None:
        """
        Save this item list collection to a Parquet file.  This supports two
        types of Parquet files: “native” collections store one row per list,
        with the item list contents in a repeated structure column named
        ``items``; this layout fully preserves the item list collection,
        including empty item lists.  The “flat” layout is easier to work with in
        software such as Pandas, but cannot store empty item lists.

        Args:
            layout:
                The table layout to use.
            batch_size:
                The Arrow record batch size.
            compression:
                The compression scheme to use.
            mkdir:
                Whether to create the parent directories if they don't exist.
        """
        if mkdir:
            Path(path).parent.mkdir(parents=True, exist_ok=True)

        if layout == "flat":
            self.to_df().to_parquet(path, compression=compression)
            return

        writer = None
        try:
            for batch in self.record_batches(batch_size):
                if writer is None:
                    writer = ParquetWriter(
                        Path(path), batch.schema, compression=compression or "snappy"
                    )
                writer.write_batch(batch)
        finally:
            if writer is not None:
                writer.close()

    @overload
    @classmethod
    def load_parquet(
        cls,
        path: PathLike[str] | list[PathLike[str]],
        *,
        layout: Literal["native"] = "native",
    ) -> ItemListCollection: ...
    @overload
    @classmethod
    def load_parquet(
        cls,
        path: PathLike[str] | list[PathLike[str]],
        key: type[K] | Sequence[Column] | Column,
        *,
        layout: Literal["flat"],
    ) -> ItemListCollection: ...
    @classmethod
    def load_parquet(
        cls,
        path: PathLike[str] | list[PathLike[str]],
        key: type[K] | Sequence[Column] | Column | None = None,
        *,
        layout: Literal["native", "flat"] = "native",
    ) -> ItemListCollection:
        """
        Load this item list from a Parquet file.

        Args:
            path:
                Path to the Parquet file to load.
            key:
                The key to use (only when loading tabular layout).
            layout:
                The layout to use, either LensKit native layout or a flat tabular layout.
        """
        if isinstance(path, list):
            path = [Path(p) for p in path]
        else:
            path = Path(path)
        dataset = ParquetDataset(path)  # type: ignore

        if layout == "native":
            if key is not None:
                raise ValueError("cannot specify key in native format")

            table = dataset.read()
            return cls.from_arrow(table)

        elif layout == "flat":
            tbl = dataset.read_pandas()

            return cls.from_df(tbl.to_pandas(), key)

        else:  # pragma: nocover
            raise ValueError(f"unsupported layout {layout}")

    def record_batches(
        self,
        batch_size: int = 5000,
        columns: dict[str, pa.DataType] | None = None,
        *,
        layout: Literal["native", "flat"] = "native",
    ) -> Generator[pa.RecordBatch, None, None]:
        """
        Get the item list collection as Arrow record batches (in native layout).
        """
        if columns is None:
            columns = self.list_schema

        for batch in chunked(self.items(), batch_size):
            keys = pa.Table.from_pylist([key_dict(k) for (k, _il) in batch])
            schema = pa.list_(pa.struct(columns))  # type: ignore
            col_elts = [
                il.to_arrow(ids=True, numbers=False, type="array", columns=columns)
                for (_k, il) in batch
            ]
            col = pa.array(col_elts, schema)
            tbl = keys.add_column(keys.num_columns, "items", col)
            if layout == "flat":
                tbl = explode_column(tbl, "items").flatten()
                tbl = tbl.rename_columns(
                    [(c[6:] if c.startswith("items.") else c) for c in tbl.column_names]
                )
            yield from tbl.to_batches()

    @property
    def key_fields(self) -> tuple[str]:
        "The names of the key fields."
        return key_fields(self.key_type)

    @property
    def key_type(self) -> type[KL]:
        """
        The type of collection keys.
        """
        return self._key_class

    @property
    @abstractmethod
    def list_schema(self) -> dict[str, pa.DataType]:
        """
        Get the schema for the lists in this ILC.
        """

    @overload
    def lookup(self, key: tuple) -> ItemList | None: ...
    @overload
    def lookup(self, *key: ID, **kwkey: ID) -> ItemList | None: ...
    @abstractmethod
    def lookup(self, *args, **kwargs) -> ItemList | None:  # pragma: nocover
        """
        Look up a list by key.  If multiple lists have the same key, this
        returns the **last** (like a dictionary).

        This method can be called with the key tuple as a single argument (and
        this can be either the actual named tuple, or an ordinary tuple of the
        same length), or with the individual key fields as positional or named
        arguments.

        Args:
            key:
                The key tuple or key tuple fields.
        """
        raise NotImplementedError()

    def lookup_projected(self, key: tuple) -> ItemList | None:
        """
        Look up an item list using a *projected* key.  A projected key is a key
        that may have additional fields beyond those defined by this collection,
        that are ignored for the purposes of lookup.

        Args:
            key:
                The key.  Must be a named tuple (e.g. a key obtained from
                another item list collection).

        Returns:
            The item list with the specified key, projected to this collection's
            key fields, or ``None`` if no such list exists.
        """
        kp = project_key(key, self._key_class)
        return self.lookup(kp)

    def total_items(self) -> int:
        """
        Count the total number of items across all lists in this collection.
        """
        return sum(len(il) for il in self.lists())

    @abstractmethod
    def items(self) -> Iterator[tuple[KL, ItemList]]:
        "Iterate over item lists and keys."
        ...

    def lists(self) -> Iterator[ItemList]:
        "Iterate over item lists without keys."
        return (il for (_k, il) in self.items())

    def keys(self) -> Iterator[KL]:
        "Iterate over keys."
        return (k for (k, _il) in self.items())

    @abstractmethod
    def __len__(self) -> int: ...

    def __iter__(self) -> Iterator[tuple[KL, ItemList]]:
        return self.items()

    @abstractmethod
    def __getitem__(self, pos: int, /) -> tuple[KL, ItemList]:
        """
        Get an item list and its key by position.

        Args:
            pos:
                The position in the list (starting from 0).

        Returns:
            The key and list at position ``pos``.

        Raises:
            IndexError:
                when ``pos`` is out-of-bounds.
        """
        pass


class ItemListCollector(Protocol):
    """
    Collect item lists with associated keys, as in :class:`ItemListCollection`.
    """

    @abstractmethod
    def add(self, list: ItemList, *fields: ID, **kwfields: ID):  # pragma: nocover
        """
        Add a single item list to this list.

        Args:
            list:
                The item list to add.
            fields, kwfields:
                The key fields for this list.
        """
        raise NotImplementedError()

    @abstractmethod
    def add_from(self, other: ItemListCollection, **fields: ID):
        """
        Add all collection from another collection to this collection.  If field
        values are supplied, they are used to supplement or overwrite the keys
        in ``other``; a common use case is to add results from multiple
        recommendation runs and save them a single field.

        Args:
            other:
                The item list collection to incorporate into this one.
            fields:
                Additional key fields (must be specified by name).
        """
        raise NotImplementedError()


class MutableItemListCollection(ItemListCollector, ItemListCollection[K], Generic[K]):
    """
    Intersection type of :class:`ItemListCollection` and
    :class:`ItemListCollector`.
    """

    pass



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

# pyright: strict
from dataclasses import dataclass
from typing import Literal

from pytest import mark

from lenskit.pipeline import Component, Pipeline, PipelineBuilder
from lenskit.pipeline.nodes import ComponentInstanceNode


@dataclass
class PrefixConfig:
    prefix: str


class Prefixer(Component[str]):
    config: PrefixConfig

    def __call__(self, msg: str) -> str:
        return self.config.prefix + msg


class Question:
    "test component that is not configurable but is a class"

    def __call__(self, msg: str) -> str:
        return msg + "?"


def exclaim(msg: str) -> str:
    return msg + "!"


def _clone(builder: PipelineBuilder, pipe: Pipeline, what: Literal["pipe", "builder"]) -> Pipeline:
    match what:
        case "pipe":
            return pipe.clone()
        case "builder":
            return builder.clone().build()


@mark.parametrize("what", ["pipe", "builder"])
def test_pipeline_clone(what: Literal["pipe", "builder"]):
    builder = PipelineBuilder()
    msg = builder.create_input("msg", str)
    builder.add_component("prefix", Prefixer, PrefixConfig(prefix="scroll named "), msg=msg)
    builder.default_component("prefix")

    pipe = builder.build()
    assert pipe.run(msg="FOOBIE BLETCH") == "scroll named FOOBIE BLETCH"
    comp = pipe.node("prefix").component  # type: ignore

    p2 = _clone(builder, pipe, what)
    n2 = p2.node("prefix")
    assert isinstance(n2, ComponentInstanceNode)
    assert isinstance(n2.component, Prefixer)
    assert n2.component is not comp
    assert n2.component.config.prefix == comp.config.prefix  # type: ignore

    assert p2.run(msg="HACKEM MUCHE") == "scroll named HACKEM MUCHE"


@mark.parametrize("what", ["pipe", "builder"])
def test_pipeline_clone_with_function(what: Literal["pipe", "builder"]):
    comp = Prefixer(prefix="scroll named ")

    builder = PipelineBuilder()
    msg = builder.create_input("msg", str)
    pfx = builder.add_component("prefix", comp, msg=msg)
    builder.add_component("exclaim", exclaim, msg=pfx)
    builder.default_component("exclaim")

    pipe = builder.build()
    assert pipe.run(msg="FOOBIE BLETCH") == "scroll named FOOBIE BLETCH!"

    p2 = _clone(builder, pipe, what)

    assert p2.run(msg="HACKEM MUCHE") == "scroll named HACKEM MUCHE!"


@mark.parametrize("what", ["pipe", "builder"])
def test_pipeline_clone_with_nonconfig_class(what: Literal["pipe", "builder"]):
    comp = Prefixer(prefix="scroll named ")

    builder = PipelineBuilder()
    msg = builder.create_input("msg", str)
    pfx = builder.add_component("prefix", comp, msg=msg)
    builder.add_component("question", Question(), msg=pfx)
    builder.default_component("question")

    pipe = builder.build()
    assert pipe.run(msg="FOOBIE BLETCH") == "scroll named FOOBIE BLETCH?"

    p2 = _clone(builder, pipe, what)

    assert p2.run(msg="HACKEM MUCHE") == "scroll named HACKEM MUCHE?"


@mark.parametrize("what", ["pipe", "builder"])
def test_clone_defaults(what: Literal["pipe", "builder"]):
    builder = PipelineBuilder()
    msg = builder.create_input("msg", str)
    builder.default_connection("msg", msg)
    builder.add_component("return", exclaim)
    builder.default_component("return")

    pipe = builder.build()
    assert pipe.run(msg="hello") == "hello!"

    p2 = _clone(builder, pipe, what)

    assert p2.run(msg="hello") == "hello!"


@mark.parametrize("what", ["pipe", "builder"])
def test_clone_alias(what: Literal["pipe", "builder"]):
    builder = PipelineBuilder()
    msg = builder.create_input("msg", str)
    excl = builder.add_component("exclaim", exclaim, msg=msg)
    builder.alias("return", excl)

    pipe = builder.build()
    assert pipe.run("return", msg="hello") == "hello!"

    p2 = _clone(builder, pipe, what)

    assert p2.run("return", msg="hello") == "hello!"


def test_clone_hash():
    pipe = PipelineBuilder()
    msg = pipe.create_input("msg", str)
    pipe.default_connection("msg", msg)
    excl = pipe.add_component("exclaim", exclaim)
    pipe.alias("return", excl)

    pipe = pipe.build()
    assert pipe.run("return", msg="hello") == "hello!"

    p2 = pipe.clone()

    assert p2.run("return", msg="hello") == "hello!"
    assert p2.config_hash == pipe.config_hash



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Tests on the ML-20M data set.
"""

import logging
from pathlib import Path

import pytest

from lenskit.basic import PopScorer
from lenskit.batch import recommend
from lenskit.data import Dataset, from_interactions_df, load_movielens
from lenskit.pipeline import topn_pipeline

_log = logging.getLogger(__name__)

_ml_path = Path("data/ml-20m.zip")


@pytest.fixture(scope="module")
def ml20m():
    if _ml_path.exists():
        return load_movielens(_ml_path)
    else:
        pytest.skip("ML-20M not available")


@pytest.mark.slow
@pytest.mark.realdata
@pytest.mark.parametrize("n_jobs", [1, 2])
def test_pop_recommend(ml20m: Dataset, rng, n_jobs):
    users = rng.choice(ml20m.users.ids(), 10000, replace=False)
    scorer = PopScorer()
    pipe = topn_pipeline(scorer)

    _log.info("training %s", pipe)
    pipe.train(ml20m)
    _log.info("recommending with %s", pipe)
    recs = recommend(pipe, users, 10, n_jobs=n_jobs)

    assert len(recs) == 10000



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import logging
import pickle
from typing import Any

import numpy as np
import pandas as pd

from pytest import approx

from lenskit.basic import BiasScorer
from lenskit.basic.composite import FallbackScorer
from lenskit.basic.history import KnownRatingScorer
from lenskit.data import Dataset
from lenskit.data.items import ItemList
from lenskit.data.types import ID
from lenskit.operations import predict, score
from lenskit.pipeline import Pipeline, PipelineBuilder, RecPipelineBuilder
from lenskit.testing import BasicComponentTests

_log = logging.getLogger(__name__)


class TestFallbackScorer(BasicComponentTests):
    component = FallbackScorer


def test_fallback_fill_missing(ml_ds: Dataset):
    pipe = PipelineBuilder()
    user = pipe.create_input("user", int)
    items = pipe.create_input("items")

    known = KnownRatingScorer()
    s1 = pipe.add_component("known", known, query=user, items=items)
    bias = BiasScorer()
    s2 = pipe.add_component("bias", bias, query=user, items=items)

    fallback = FallbackScorer()
    score = pipe.add_component("mix", fallback, primary=s1, backup=s2)

    pipe = pipe.build()
    pipe.train(ml_ds)

    # the first 2 of these are rated, the 3rd does not exist, and the other 2 are not rated
    items = [50, 17, 210, 1172, 2455]
    scored = pipe.run(score, user=2, items=ItemList(item_ids=items))

    assert len(scored) == len(items)
    assert np.all(scored.ids() == items)
    scores = scored.scores()
    assert scores is not None
    assert not np.any(np.isnan(scored.scores()))

    assert scores[:2] == approx(known(2, ItemList(item_ids=items[:2])).scores())
    assert scores[2:] == approx(bias(2, ItemList(item_ids=items[2:])).scores())


def test_fallback_double_bias(rng: np.random.Generator, ml_ds: Dataset):
    builder = RecPipelineBuilder()
    builder.scorer(BiasScorer(damping=50))
    builder.predicts_ratings(fallback=BiasScorer(damping=0))
    pipe = builder.build("double-bias")

    _log.info("pipeline configuration: %s", pipe.config.model_dump_json(indent=2))

    pipe.train(ml_ds)

    for user in rng.choice(ml_ds.users.ids(), 100):
        items = rng.choice(ml_ds.items.ids(), 500)
        scores = score(pipe, user, items)
        scores = scores.scores()
        assert scores is not None
        assert not np.any(np.isnan(scores))

        preds = predict(pipe, user, items)

        preds = preds.scores()
        assert preds is not None
        assert not np.any(np.isnan(preds))

        assert scores == approx(preds)



from . import (
    als,
    basic,
    batch,
    config,
    data,
    flexmf,
    funksvd,
    hpf,
    knn,
    logging,
    metrics,
    operations,
    pipeline,
    sklearn,
    splitting,
    stats,
    torch,
    training,
)
from .config import configure, lenskit_config
from .data import Dataset, DatasetBuilder
from .operations import predict, recommend, score
from .pipeline import Component, Pipeline, RecPipelineBuilder, topn_pipeline
from .splitting import TTSplit

__all__ = [
    # modules
    "batch",
    "config",
    "data",
    "logging",
    "metrics",
    "pipeline",
    "operations",
    "splitting",
    "stats",
    "torch",
    "training",
    # component modules
    "als",
    "basic",
    "flexmf",
    "funksvd",
    "hpf",
    "knn",
    "sklearn",
    # setup
    "configure",
    "lenskit_config",
    # operations
    "predict",
    "recommend",
    "score",
    # pipeline
    "Pipeline",
    "Component",
    "RecPipelineBuilder",
    "topn_pipeline",
    # data
    "Dataset",
    "DatasetBuilder",
    # splitting
    "TTSplit",
]

__version__: str



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import os
import platform
import re
import sys
from dataclasses import dataclass
from importlib.metadata import distributions, version
from pathlib import Path

import click
import psutil
import threadpoolctl
from cpuinfo import get_cpu_info
from humanize import metric, naturalsize
from rich.console import Console, ConsoleOptions, group
from rich.padding import Padding
from rich.table import Table

from lenskit import __version__, _accel
from lenskit.logging import get_logger, stdout_console
from lenskit.parallel import ensure_parallel_init
from lenskit.parallel.config import (
    ParallelConfig,
    effective_cpu_count,
    get_parallel_config,
    subprocess_config,
)
from lenskit.parallel.ray import ray_available

_log = get_logger(__name__)
_gh_out: Path | None = None


@click.command("doctor")
@click.option(
    "--github-output",
    "gh_output",
    envvar="GITHUB_OUTPUT",
    type=Path,
    help="Path to GitHub Actions output file.",
)
@click.option("--packages/--no-packages", default=False, help="List installed packages.")
@click.option("--paths/--no-paths", default=False, help="List search paths.")
@click.option("--full", is_flag=True, default=False, help="Show all output.")
def doctor(gh_output: Path | None, packages: bool, paths: bool, full: bool):
    """
    Inspect installed LensKit version and environment.
    """
    global _gh_out
    _gh_out = gh_output
    ensure_parallel_init()
    console = stdout_console()
    console.print(inspect_version(), highlight=False)
    console.print(inspect_platform(), highlight=False)
    console.print(inspect_system(), highlight=False)
    console.print(inspect_compute(), highlight=False)
    if ray_available():
        console.print(inspect_ray(), highlight=False)
    console.print(inspect_parallel_config(), highlight=False)
    console.print(inspect_env(paths or full), highlight=False)
    if packages or full:
        console.print(inspect_packages(), highlight=False)


@dataclass
class kvp:
    name: str
    value: str | int | bool | float | None
    level: int = 1
    value_style: str = "cyan"

    def __rich_console__(self, console: Console, options: ConsoleOptions):
        text = ""
        if self.level == 1:
            text += f"[bold]{self.name}[/bold]: "
        else:
            text += "  " * (self.level - 1)
            text += self.name + ": "
        text += f"[{self.value_style}]"
        text += str(self.value)
        text += f"[/{self.value_style}]"
        return [text]


@group()
def inspect_version():
    dist_ver = version("lenskit")
    if _gh_out:
        with _gh_out.open("at") as ghf:
            print(f"lenskit_version={dist_ver}", file=ghf)

    yield f"[bold]LensKit version:[/bold] [cyan]{dist_ver}[/cyan]"
    if str(dist_ver) != __version__:
        yield f"   [yellow]Version mismatch, internal package version is {__version__}[/yellow]"


@group()
def inspect_platform():
    yield kvp("Python version", platform.python_version())
    yield kvp("Platform", platform.platform(), level=2)
    yield kvp("Location", sys.executable, level=2)


@group()
def inspect_system():
    yield ""
    yield "[bold]System information:[/bold]"

    cpu = get_cpu_info()
    yield kvp("Processor", cpu["brand_raw"], level=2)
    if freq := cpu.get("hz_advertised", None):
        yield kvp("CPU Frequency", metric(freq[0], unit="Hz"), level=2)

    eff_cpu = effective_cpu_count()
    ncpus = os.cpu_count()
    cpus = f"[bold]{ncpus}[/bold]"
    nphys = psutil.cpu_count(logical=False)
    if nphys != ncpus:
        cpus += f" ({nphys} physical)"
    if ncpus != eff_cpu:
        cpus += f", limited to {ncpus}"
    yield kvp("CPU cores", cpus, level=2)

    vmem = psutil.virtual_memory()
    yield kvp(
        "Memory",
        f"[bold]{naturalsize(vmem.total, binary=True)}[/bold]"
        f" ({naturalsize(vmem.available, binary=True)} available)",
        level=2,
    )


@group()
def inspect_parallel_config():
    yield ""
    yield "[bold]Parallel configuration[/bold]:"
    yield "  [cyan]Main configuration[/cyan]:"
    pc = get_parallel_config()
    yield _inspect_pc(pc, level=3)
    pc = subprocess_config()
    yield "  [blue]Default worker configuration[/blue]:"
    yield _inspect_pc(pc, level=3)


@group()
def _inspect_pc(pc: ParallelConfig, *, level: int):
    yield kvp("processes", pc.processes, level=level)
    yield kvp("threads", pc.threads, level=level)
    yield kvp("backend threads", pc.backend_threads, level=level)


@group()
def inspect_compute():
    import numpy as np
    import torch

    try:
        import cupy  # type: ignore

        _log.debug("imported CuPy version %s", cupy.__version__)
    except ImportError:
        _log.debug("CuPy unavailable")
        cupy = None

    yield ""
    yield kvp("NumPy version", np.__version__)
    yield kvp("PyTorch version", torch.__version__)
    if _gh_out:
        with _gh_out.open("at") as ghf:
            print(f"numpy_version={np.__version__}", file=ghf)
            print(f"pytorch_version={torch.__version__}", file=ghf)

    yield "[bold]PyTorch backends[/bold]:"
    yield kvp("cpu", torch.backends.cpu.get_cpu_capability(), level=2)
    for mod in [torch.cuda, torch.backends.mkl, torch.backends.mps]:
        if mod.is_available():
            stat = "available"
        elif hasattr(mod, "is_built") and mod.is_built():
            stat = "unavailable"
        else:
            stat = "absent"
        name = mod.__name__.split(".")[-1]
        yield kvp(name, stat, level=2)

    if torch.cuda.is_available():
        yield ""
        yield "[bold]PyTorch GPUs[/bold]:"
        for dev in range(torch.cuda.device_count()):
            props = torch.cuda.get_device_properties(dev)
            yield "  [green]cuda:{}[/green]: [bold cyan]{}[/bold cyan]".format(dev, props.name)
            yield kvp("capability", f"{props.major}.{props.minor}", level=2)
            yield kvp("memory", naturalsize(props.total_memory, binary=True), level=2)
            yield kvp("L2 cache", naturalsize(props.L2_cache_size, binary=True), level=2)
            yield kvp("MP count", props.multi_processor_count, level=2)
            if cupy is not None:
                cd = cupy.cuda.Device(dev)
                yield kvp("warp size", cd.attributes["WarpSize"], level=2)
                yield kvp("blocks/MP", cd.attributes["MaxBlocksPerMultiprocessor"], level=2)
                yield kvp("clock rate", metric(cd.attributes["ClockRate"], unit="Hz"), level=2)
                yield kvp("sp/dp ratio", cd.attributes["SingleToDoublePrecisionPerfRatio"], level=2)

    yield ""
    yield "[bold]Threading layers[/bold]:"
    yield "  Rayon:"
    yield kvp("threads", _accel.thread_count(), level=3)
    for i, pool in enumerate(threadpoolctl.threadpool_info(), 1):
        yield f"  Backend {i}:"
        for k, v in pool.items():
            yield kvp(k, v, level=3)


@group()
def inspect_ray():
    import ray  # type: ignore

    yield ""
    yield "[bold]Ray cluster ([yellow]experimental[/yellow])[/bold]:"

    try:
        ray.init("auto", configure_logging=False)
    except ConnectionError:
        yield "  Installed but inactive"
    except RuntimeError:
        yield "  Cannot connect"
    else:
        yield "  Resources:"
        for name, val in ray.cluster_resources().items():
            if name.startswith("node:"):
                continue
            if name.endswith("memory"):
                val = naturalsize(val)
            yield kvp(name, val, level=3)


@group()
def inspect_env(paths: bool):
    yield ""
    yield "[bold]Relevant environment variables[/bold]:"
    for k, v in os.environ.items():
        if re.match(r"^(LK_|OMP_|NUMBA_|MKL_|TORCH_|PY)", k):
            yield kvp(k, v, level=2)

    if not paths:
        return

    yield ""
    yield "[bold]Python search paths[/bold]:"
    for path in sys.path:
        yield f"- {path}"

    yield ""
    yield "[bold]Executable search paths[/bold]:"
    exe_paths = os.environ["PATH"].split(os.pathsep)
    for path in exe_paths:
        yield f"- {path}"


def inspect_packages():
    dists = sorted(distributions(), key=lambda d: d.name or "UNNAMED")

    n = len(dists)
    table = Table(title=f"Installed Packages ({n})")
    table.add_column("Package")
    table.add_column("Version", justify="right")

    for dist in dists:
        table.add_row(dist.name, dist.version)

    return Padding(table, (1, 0, 0, 2))



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from __future__ import annotations

from datetime import datetime
from typing import Literal

import numpy as np
import pandas as pd
from pydantic import BaseModel
from typing_extensions import override

from lenskit.data import Dataset, ItemList, Vocabulary
from lenskit.logging import get_logger
from lenskit.pipeline import Component
from lenskit.training import Trainable, TrainingOptions

_log = get_logger(__name__)


class PopConfig(BaseModel):
    """
    Configuration for popularity scoring.
    """

    score: Literal["quantile", "rank", "count"] = "quantile"
    """
    The method for computing popularity scores.  For all methods, higher scores
    represent more popular items.
    """


class PopScorer(Component[ItemList], Trainable):
    """
    Score items by their popularity.  Use with :py:class:`TopN` to get a
    most-popular-items recommender.

    Stability:
        Caller

    Attributes:
        item_pop_:
            Item popularity scores.
    """

    config: PopConfig

    items_: Vocabulary
    item_scores_: np.ndarray[tuple[int], np.dtype[np.float32]]

    @override
    def train(self, data: Dataset, options: TrainingOptions = TrainingOptions()):
        if hasattr(self, "item_scores_") and not options.retrain:
            return

        _log.info("counting item popularity")
        self.items_ = data.items
        stats = data.item_stats()
        scores = stats["count"]
        self.item_scores_ = np.require(self._train_internal(scores).values, np.float32)

    def _train_internal(self, scores: pd.Series) -> pd.Series:
        if self.config.score == "rank":
            _log.info("ranking %d items", len(scores))
            return scores.rank()
        elif self.config.score == "quantile":
            _log.info("computing quantiles for %d items", len(scores))
            cmass = scores.sort_values()
            cmass = cmass.cumsum()
            cdens = cmass / scores.sum()
            return cdens.reindex(scores.index)
        elif self.config.score == "count":
            _log.info("scoring items with their rating counts")
            return scores
        else:
            raise ValueError("invalid scoring method " + repr(self.config.score))

    def __call__(self, items: ItemList) -> ItemList:
        inums = items.numbers(vocabulary=self.items_, missing="negative")
        mask = inums >= 0
        good_inums = inums[mask]
        _log.debug("getting popularity scores", n_good=len(good_inums), n_all=len(inums))
        scores = np.full(len(items), np.nan, np.float32)
        scores[mask] = self.item_scores_[inums[mask]]
        return ItemList(items, scores=scores)


class TimeBoundedPopConfig(PopConfig):
    cutoff: datetime
    """
    Time window for computing popularity scores.
    """


class TimeBoundedPopScore(PopScorer):
    """
    Score items by their time-bounded popularity, i.e., the popularity in the
    most recent `time_window` period.  Use with :py:class:`TopN` to get a
    most-popular-recent-items recommender.

    Attributes:
        item_scores_(pandas.Series):
            Time-bounded item popularity scores.
    """

    config: TimeBoundedPopConfig

    @override
    def train(self, data: Dataset, options: TrainingOptions = TrainingOptions()):
        if hasattr(self, "item_scores_") and not options.retrain:
            return

        _log.info("counting time-bounded item popularity")
        log = data.interaction_table(format="pandas", original_ids=True)

        item_scores = None
        if "timestamp" not in log.columns:
            _log.warning("no timestamps in interaction log; falling back to PopScorer")
            super().train(data, options)
            return
        else:
            item_ids = log["item_id"][log["timestamp"] > self.config.cutoff]
            counts = item_ids.value_counts().reindex(data.items.index, fill_value=0)

            item_scores = super()._train_internal(counts)

        self.items_ = data.items
        self.item_scores_ = np.require(item_scores.reindex(self.items_.ids()).values, np.float32)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from pathlib import Path

from numpy.random import Generator, SeedSequence

from hypothesis import assume, given
from hypothesis import strategies as st
from pytest import skip

from lenskit.data.query import RecQuery
from lenskit.random import derivable_rng, load_seed, make_seed


@given(
    st.one_of(
        st.integers(min_value=0),
        st.binary(),
        st.uuids(),
        st.text(),
        st.lists(st.integers(min_value=0)),
    )
)
def test_make_seed(seed):
    ss = make_seed(seed)
    assert isinstance(ss, SeedSequence)
    state = ss.generate_state(1)
    assert len(state) == 1
    assert state[0]


@given(st.integers(min_value=0), st.emails(), st.emails())
def test_derivable_fixed(seed: int, uid1: str, uid2: str):
    assume(uid1 != uid2)

    drng = derivable_rng(seed)

    v1a = drng(RecQuery.create(uid1)).integers(1_000_000)
    v1b = drng(RecQuery.create(uid1)).integers(1_000_000)
    v2 = drng(RecQuery.create(uid2)).integers(1_000_000)

    assert v1a != v2
    assert v1a != v1b


@given(st.integers(min_value=0), st.emails(), st.emails())
def test_derivable_user(seed: int, uid1: str, uid2: str):
    assume(uid1 != uid2)

    drng = derivable_rng((seed, "user"))

    v1a = drng(RecQuery.create(uid1)).integers(1_000_000)
    v1b = drng(RecQuery.create(uid1)).integers(1_000_000)
    v2 = drng(RecQuery.create(uid2)).integers(1_000_000)

    assert v1a != v2
    assert v1a == v1b

    vnone = drng(None).integers(1_000_000)
    assert vnone != v1a
    assert vnone != v2


def test_load_seed_json():
    file = Path(__file__).parent / "config.json"
    seed = load_seed(file)
    assert seed.entropy == [20250205]


def test_load_seed_yaml():
    try:
        import yaml
    except ImportError:
        skip("pyyaml not installed")

    file = Path(__file__).parent / "config.yaml"
    seed = load_seed(file, "random_seed")
    assert seed.entropy == [20250205]


def test_load_seed_toml():
    file = Path(__file__).parent / "config.toml"
    seed = load_seed(file, "rng.seed")
    assert seed.entropy == [20250201]



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import os
from pathlib import Path

import pandas as pd
import pyarrow as pa
import pyarrow.compute as pc

from pytest import mark

from lenskit.data import load_amazon_ratings
from lenskit.data.vocab import Vocabulary

AZ_NAME = os.environ.get("LK_TEST_AZ_SET", "Video_Games")

AZ14_TEST = Path(f"data/az14/ratings_{AZ_NAME}.csv")
AZ18_TEST = Path(f"data/az18/{AZ_NAME}.csv")
AZ23_TEST = Path(f"data/az23/{AZ_NAME}.csv.gz")


@mark.skipif(not AZ14_TEST.exists(), reason="input data does not exist")
@mark.xfail(reason="duplicate ratings")
@mark.realdata
def test_amazon_2014():
    df = pd.read_csv(AZ14_TEST, names=["user_id", "item_id", "rating", "timestamp"])
    data = load_amazon_ratings(AZ14_TEST)

    assert data.interaction_count == len(df)
    assert data.user_count == df["user_id"].nunique()
    assert data.item_count == df["item_id"].nunique()


@mark.skipif(not AZ18_TEST.exists(), reason="input data does not exist")
@mark.xfail(reason="duplicate ratings")
@mark.realdata
def test_amazon_2018():
    df = pd.read_csv(AZ18_TEST, names=["item_id", "user_id", "rating", "timestamp"])
    data = load_amazon_ratings(AZ18_TEST)

    assert data.interaction_count == len(df)
    assert data.user_count == df["user_id"].nunique()
    assert data.item_count == df["item_id"].nunique()


@mark.skipif(not AZ23_TEST.exists(), reason="input data does not exist")
@mark.realdata
def test_amazon_2023_vocab():
    "targeted vocab tests for broken data set"
    df = pd.read_csv(AZ23_TEST)

    user_ids = df["user_id"].unique()
    item_ids = df["parent_asin"].unique()

    user_vocab = Vocabulary(user_ids)
    unos = user_vocab.numbers(df["user_id"], format="arrow", missing="null")
    assert unos.null_count == 0
    assert pc.count_distinct(unos).as_py() == len(user_ids)

    item_vocab = Vocabulary(item_ids)
    inos = item_vocab.numbers(df["parent_asin"], format="arrow", missing="null")
    assert inos.null_count == 0
    assert pc.count_distinct(inos).as_py() == len(item_ids)


@mark.skipif(not AZ23_TEST.exists(), reason="input data does not exist")
@mark.realdata
def test_amazon_2023():
    df = pd.read_csv(AZ23_TEST)
    data = load_amazon_ratings(AZ23_TEST)

    assert data.interaction_count == len(df)
    assert data.user_count == df["user_id"].nunique()
    assert data.item_count == df["parent_asin"].nunique()



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from pytest import approx

from lenskit.data import ItemList
from lenskit.metrics import GeometricRankWeight
from lenskit.metrics.reranking import rank_biased_overlap


def test_rbo_empty_reference():
    ref = ItemList()
    rerank = ItemList([1, 2, 3], ordered=True)
    assert rank_biased_overlap(ref, rerank) == 0.0


def test_rbo_empty_rerank():
    ref = ItemList([1, 2, 3], ordered=True)
    rerank = ItemList()
    assert rank_biased_overlap(ref, rerank) == 0.0


def test_rbo_perfect_match():
    ref = ItemList([1, 2, 3, 4, 5], ordered=True)
    rerank = ItemList([1, 2, 3, 4, 5], ordered=True)
    weight = GeometricRankWeight(0.9)
    score = rank_biased_overlap(ref, rerank, k=5, weight=weight)
    assert score > 0.89
    assert score <= 1.0


def test_rbo_no_overlap():
    ref = ItemList([1, 2, 3], ordered=True)
    rerank = ItemList([4, 5, 6], ordered=True)
    weight = GeometricRankWeight(0.9)
    assert rank_biased_overlap(ref, rerank, k=3, weight=weight) == approx(0.0)


def test_rbo_partial_overlap():
    ref = ItemList([1, 2, 3, 4, 5], ordered=True)
    rerank = ItemList([3, 2, 6, 7, 1], ordered=True)
    weight = GeometricRankWeight(0.9)
    score = rank_biased_overlap(ref, rerank, k=5, weight=weight)
    # RBO should be between 0 and 1
    assert 0 < score < 1


def test_rbo_k_smaller_than_list():
    ref = ItemList([1, 2, 3, 4, 5], ordered=True)
    rerank = ItemList([5, 4, 3, 2, 1], ordered=True)
    weight = GeometricRankWeight(0.9)
    # consider only top 3
    score = rank_biased_overlap(ref, rerank, k=3, weight=weight)
    assert 0 <= score <= 1



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import warnings
from collections.abc import Iterator, Mapping, Sequence
from typing import Any, Generic, overload

import pandas as pd
import pyarrow as pa

from lenskit.diagnostics import DataWarning
from lenskit.logging import get_logger

from ..adapt import Column, column_name, normalize_columns
from ..items import ItemList
from ._base import ItemListCollection, MutableItemListCollection
from ._keys import ID, GenericKey, K, Ko, create_key_type, key_dict, key_fields

_log = get_logger(__name__)


class ListILC(MutableItemListCollection[K], Generic[K]):
    """
    Mutable item list collection backed by a Python list.
    """

    _key_class: type[K]
    _lists: list[tuple[K, ItemList]]
    _index: dict[K, int] | None = None
    _list_schema: dict[str, pa.DataType]

    def __init__(self, key: type[K] | Sequence[str], *, index: bool = True):
        """
        Create a new item list collection.
        """
        if isinstance(key, type):
            self._key_class = key
        else:
            self._key_class = create_key_type(*key)  # type: ignore

        self._lists = []
        if index:
            self._index = {}
        self._list_schema = {}

    @property
    def list_schema(self):
        return self._list_schema

    @overload
    @classmethod
    def from_dict(
        cls, data: Mapping[GenericKey | ID, ItemList], key: type[K]
    ) -> ItemListCollection[K]: ...
    @overload
    @classmethod
    def from_dict(
        cls, data: Mapping[GenericKey | ID, ItemList], key: Sequence[str] | str | None = None
    ) -> ItemListCollection[GenericKey]: ...
    @classmethod
    def from_dict(
        cls,
        data: Mapping[GenericKey | ID, ItemList],
        key: type[K] | Sequence[str] | str | None = None,
    ) -> ItemListCollection[Any]:
        """
        Create an item list collection from a dictionary.
        """
        if key is not None:
            if isinstance(key, str):
                key = (key,)
            ilc = ListILC(key)
        else:
            k = next(iter(data.keys()))
            if isinstance(k, tuple) and getattr(k, "_fields"):
                ilc = ListILC(type(k))
            else:
                warnings.warn(
                    "no key specified but data does not use named tuples, using default field 'id'",
                    DataWarning,
                )
                ilc = ListILC(["id"])

        for k, il in data.items():
            if isinstance(k, tuple):
                ilc.add(il, *k)
            else:
                ilc.add(il, k)

        return ilc

    @classmethod
    def from_df(
        cls,
        df: pd.DataFrame,
        key: type[K] | Sequence[Column] | Column | None = None,
        *others: Column,
    ):
        """
        Create an item list collection from a data frame.

        .. note::

            Keys with empty item lists will be silently excluded from the output
            data.

        Args:
            df:
                The data frame to convert.
            key:
                The key type or field(s).  Can be specified as a single column
                name (or :class:`~lenskit.data.types.AliasedColumn`).
            others:
                Other columns to consider; primarily used to pass additional
                aliased columns to normalize other clumnes like the item ID.
        """
        if isinstance(key, type):
            fields = key_fields(key)
            columns = fields + others
        else:
            if key is None:
                warnings.warn(
                    "no key specified, inferring from _id columns", DataWarning, stacklevel=2
                )
                key = [n for n in df.columns if n.endswith("_id") and n != "item_id"]
            elif isinstance(key, Column):
                key = [key]
            columns = tuple(key) + others
            fields = [column_name(c) for c in key]
            key = create_key_type(*fields)  # type: ignore

        _log.debug("converting ILC from dataframe", key=key, rows=len(df))

        df = normalize_columns(df, *columns)
        ilc = cls(key)  # type: ignore
        for k, gdf in df.groupby(list(fields)):
            ilc.add(ItemList.from_df(gdf), *k)

        return ilc

    @staticmethod
    def from_arrow(table: pa.Table) -> MutableItemListCollection[Any]:
        """
        Convert an Arrow table into an item list collection.
        """
        from ._list import ListILC

        keys = table.drop("items")
        lists = table.column("items")
        ilc = ListILC(keys.schema.names)
        for i, k in enumerate(keys.to_pylist()):
            il_data = lists[i].values
            ilc.add(ItemList.from_arrow(il_data), **k)

        return ilc

    def add(self, list: ItemList, *fields: ID, **kwfields: ID):
        """
        Add a single item list to this list.

        Args:
            list:
                The item list to add.
            fields, kwfields:
                The key fields for this list.
        """
        key = self._key_class(*fields, **kwfields)  # type: ignore
        self._add(key, list)

    def add_from(self, other: ItemListCollection[Ko], **fields: ID):
        """
        Add all collection from another collection to this collection.  If field
        values are supplied, they are used to supplement or overwrite the keys
        in ``other``; a common use case is to add results from multiple
        recommendation runs and save them a single field.

        Args:
            other:
                The item list collection to incorporate into this one.
            fields:
                Additional key fields (must be specified by name).
        """
        for key, list in other:
            if fields:
                cf = key_dict(key) | fields
                key = self._key_class(**cf)
            self._add(key, list)  # type: ignore

    def _add(self, key: K, list: ItemList):
        self._lists.append((key, list))
        if self._index is not None:
            self._index[key] = len(self._lists) - 1
        if len(list):
            for fn, ft in list.arrow_types().items():
                pft = self._list_schema.get(fn, None)
                if pft is None:
                    self._list_schema[fn] = ft
                elif not ft.equals(pft):
                    raise TypeError(f"incompatible item lists: field {fn} type {ft} != {pft}")

    @overload
    def lookup(self, key: tuple) -> ItemList | None: ...
    @overload
    def lookup(self, *key: ID, **kwkey: ID) -> ItemList | None: ...
    def lookup(self, *args, **kwargs) -> ItemList | None:
        """
        Look up a list by key.  If multiple lists have the same key, this
        returns the **last** (like a dictionary).

        This method can be called with the key tuple as a single argument (and
        this can be either the actual named tuple, or an ordinary tuple of the
        same length), or with the individual key fields as positional or named
        arguments.

        Args:
            key:
                The key tuple or key tuple fields.
        """
        if self._index is None:
            raise TypeError("cannot lookup on non-indexed collection")
        if len(args) != 1 or not isinstance(args[0], tuple):
            key = self._key_class(*args, **kwargs)
        else:
            key = args[0]

        try:
            return self._lists[self._index[key]][1]  # type: ignore
        except KeyError:
            return None

    def items(self) -> Iterator[tuple[K, ItemList]]:
        "Iterate over item lists and keys."
        return iter(self._lists)

    def __len__(self):
        return len(self._lists)

    def __getitem__(self, pos: int, /) -> tuple[K, ItemList]:
        return self._lists[pos]



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import doctest
import sys
from os import fspath
from pathlib import Path

from packaging.version import Version

from lenskit._version import lenskit_version

sys.path.append(str((Path(__file__).parent / "_ext").resolve()))

project = "LensKit"
copyright = "2018–2025 Drexel University, Boise State University, and collaborators"
author = "Michael D. Ekstrand"

release = lenskit_version()
_parsed_ver = Version(release)
version = _parsed_ver.base_version

extensions = [
    "myst_nb",
    "sphinx.ext.napoleon",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.doctest",
    "sphinx.ext.intersphinx",
    "sphinx.ext.mathjax",
    "sphinx.ext.extlinks",
    "sphinx.ext.todo",
    "sphinx_togglebutton",
    "sphinxext.opengraph",
    "sphinxcontrib.bibtex",
    "sphinxcontrib.mermaid",
    "sphinx_copybutton",
    "sphinx_new_tab_link",
    "lk_stability",
]

# set up our filenames
# source_suffix = {".rst": "restructuredtext"}
exclude_patterns = [
    "_build",
    "_ext",
    "Thumbs.db",
    ".DS_Store",
    "old/*",
]
nb_execution_mode = "off"

# layout and setup options
pygments_style = "sphinx"
highlight_language = "python3"

html_theme = "pydata_sphinx_theme"
html_logo = "LKLogo2.png"
if _parsed_ver.is_devrelease:
    html_baseurl = "https://lenskit.org/latest/"
else:
    html_baseurl = "https://lenskit.org/stable/"
html_css_files = [
    "css/custom.css",
]

html_theme_options = {
    "switcher": {
        "json_url": "https://lenskit.org/versions.json",
        "version_match": "2024.0dev",
    },
    "show_version_warning_banner": True,
    "icon_links": [
        {
            "name": "GitHub",
            "url": "https://github.com/lenskit/lkpy",
            "icon": "fa-brands fa-github",
            "type": "fontawesome",
        },
        {
            "name": "Mastodon",
            "url": "https://recsys.social/@LensKit",
            "icon": "fa-brands fa-mastodon",
            "type": "fontawesome",
        },
        {
            "name": "BlueSky",
            "url": "https://bsky.app/profile/lenskit.org",
            "icon": "fa-brands fa-bluesky",
            "type": "fontawesome",
        },
    ],
    "navbar_end": ["version-switcher", "theme-switcher", "navbar-icon-links"],
    "footer_start": ["copyright", "version", "disclaimer", "counter"],
    "footer_end": [],
    # "github_user": "lenskit",
    # 'github_repo': 'lkpy',
    # 'travis_button': False,
    # 'canonical_url': 'https://lenskit.org/',
    # 'font_family': 'Charter, serif'
    # 'font_family': '"Source Sans Pro", "Georgia Pro", Georgia, serif',
    # 'font_size': '15px',
    # 'head_font_family': '"Merriweather Sans", "Arial", sans-serif',
    # 'code_font_size': '1em',
    # 'code_font_family': '"Source Code Pro", "Consolas", "Menlo", sans-serif'
}

templates_path = ["_templates"]
html_static_path = ["_static"]

# how do we want to set up documentation?
autodoc_default_options = {"members": True, "member-order": "bysource", "show-inheritance": True}
autodoc_typehints = "description"
autodoc_type_aliases = {
    "ArrayLike": "numpy.typing.ArrayLike",
    "SeedLike": "lenskit.random.SeedLike",
    "RNGLike": "lenskit.random.RNGLike",
    "RNGInput": "lenskit.random.RNGInput",
    "IDSequence": "lenskit.data.types.IDSequence",
}
# autosummary_generate_overwrite = False
autosummary_imported_members = False
autosummary_ignore_module_all = True

# customize doc parsing
napoleon_custom_sections = [("Stability", "returns_style")]

nitpicky = True
todo_include_todos = True

# Cross-linking and external references
intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "pandas": ("http://pandas.pydata.org/pandas-docs/stable/", None),
    "pyarrow": ("https://arrow.apache.org/docs/", None),
    "numpy": ("https://numpy.org/doc/stable/", None),
    "scipy": ("https://docs.scipy.org/doc/scipy/", None),
    "sklearn": ("https://scikit-learn.org/stable/", None),
    "structlog": ("https://www.structlog.org/en/stable/", None),
    "torch": ("https://docs.pytorch.org/docs/stable/", None),
    "implicit": ("https://benfred.github.io/implicit/", None),
    "pydantic": ("https://docs.pydantic.dev/latest/", None),
    "ray": ("https://docs.ray.io/en/latest/", None),
    "torch_geometric": ("https://pytorch-geometric.readthedocs.io/en/latest/", None),
}

bibtex_bibfiles = ["lenskit.bib"]
nb_execution_mode = "off"
doctest_path = [fspath((Path(__file__).parent / "guide" / "examples").resolve())]
doctest_default_flags = (
    doctest.ELLIPSIS | doctest.IGNORE_EXCEPTION_DETAIL | doctest.NORMALIZE_WHITESPACE
)

mermaid_d3_zoom = True

# -- external links

extlinks = {
    "issue": ("https://github.com/lenskit/lkpy/issues/%s", "🐞 %s"),
    "pr": ("https://github.com/lenskit/lkpy/pull/%s", "⛙ %s"),
    "user": ("https://github.com/%s", "@%s"),
}



from __future__ import annotations

import numpy as np
import pyarrow as pa

from lenskit.data.matrix import SparseRowArray
from lenskit.data.types import ID

def transpose_csr(
    matrix: SparseRowArray, permute: bool
) -> tuple[pa.Array, pa.Int32Array, pa.Array | None]: ...
def is_sorted_coo(data: list[pa.RecordBatch], c1: str, c2: str) -> bool: ...
def argsort_descending(data: pa.Array) -> pa.Int32Array: ...
def negative_mask(n: int, indices: pa.Int32Array) -> pa.BooleanArray:
    """
    Efficiently create a boolean array that is true everywhere except ``indices``.
    """
    ...

def sample_negatives(
    coords: CoordinateTable,
    rows: np.ndarray[tuple[int], np.dtype[np.int32]],
    n_cols: int,
    *,
    n: int = 1,
    max_attempts: int = 10,
    pop_weighted: bool = False,
    seed: int,
) -> np.ndarray[tuple[int, int], np.dtype[np.int32]]: ...
def hash_array(arr: pa.Array, /) -> str: ...

class IDIndex:
    """
    Backend implementation for ID indexes / vocabularies.
    """
    def __init__(self, ids: pa.Array | None = None): ...
    def get_index(self, id: ID): ...
    def get_indexes(self, ids: pa.Array): ...
    def id_array(self) -> pa.Array: ...
    def __len__(self) -> int: ...

class CoordinateTable:
    def __init__(self, dims: int): ...
    def copy(self) -> CoordinateTable: ...
    def dimensions(self) -> int: ...
    def __len__(self) -> int: ...
    def entry_count(self) -> int: ...
    def unique_count(self) -> int: ...
    def extend(self, batch: pa.RecordBatch | list[pa.RecordBatch]) -> tuple[int, int]: ...
    def contains(self, *coords: int) -> bool: ...
    def find(self, *coords: int) -> int | None: ...
    def _get_coord(self, dim: int, position: int) -> int: ...



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

# extra comment to make license headers happy
def field_format(name: str, fs: str | None):
    if fs:
        return "{%s:%s}" % (name, fs)
    else:
        return "{%s}" % (name,)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import numpy as np
import pyarrow as pa
import pyarrow.compute as pc


def test_null_mixup(rng: np.random.Generator):
    """
    Test for Arrow bug 47234.
    """
    nums = rng.integers(0, 10_000, size=100)
    mask = np.zeros(100, dtype=np.bool_)
    mask[-1] = True

    arr = pa.array(nums, mask=mask)
    print(arr)
    assert arr.null_count == 1

    a2 = pc.fill_null(arr, -1)
    print(a2)
    assert a2.null_count == 0

    npa2 = a2.to_numpy()
    assert npa2[-1] < 0
    assert np.all(npa2[:-1] >= 0)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Extended logger providing TRACE support.
"""

from __future__ import annotations

import logging
import os
from typing import Any, Literal

import structlog
from structlog.stdlib import BoundLogger

from ._proxy import get_logger

__trace_debug = os.environ.get("LK_TRACE", "no").lower() == "debug"
# default based on env var, so that LK_TRACE=debug enabels even without
# debug config — useful for tests.
_tracing_active: bool | Literal["debug"] = "debug" if __trace_debug else False


def tracing_active() -> bool:
    """
    Query whether tracing is active.
    """
    return bool(_tracing_active)


def activate_tracing(active: bool | Literal["debug"] = True) -> None:
    """
    Mark tracing as active (or inactive).

    Global tracing state is just used to short-cut tracing.  This method should
    only be called from :class:`~lenskit.logging.LoggingConfig`.

    Args:
        active:
            The global tracing state.  If ``"debug"``, trace messages are
            emitted at DEBUG level.
    """
    global _tracing_active
    _tracing_active = active


def trace(logger: BoundLogger, *args: Any, **kwargs: Any):
    """
    Emit a trace-level message, if LensKit tracing is enabled.  Trace-level
    messages are more fine-grained than debug-level messages, and you usually
    don't want them.

    This function does not work on the lazy proxies returned by
    :func:`get_logger` and similar — it only works on bound loggers.

    Stability:
        Caller
    """
    if _tracing_active == "debug":
        logger.debug(*args, **kwargs)
    elif _tracing_active:
        meth = getattr(logger, "trace", None)
        if meth is not None:
            meth(*args, **kwargs)


def get_tracer(logger: str | BoundLogger, **initial_values: Any):
    """
    Get a tracer for efficient low-level tracing of computations.

    Stability:
        Experimental
    """
    if isinstance(logger, str):
        logger = get_logger(logger)
    if initial_values:
        logger = logger.bind(**initial_values)

    if _tracing_active:
        logger.debug("creating active tracer")
        return ActiveTracer(logger)
    else:
        logger.debug("tracing deactivated, creating no-op tracer")
        return Tracer(logger)


class Tracer:
    """
    Logger-like thing that is only for TRACE-level events.

    This class is designed to support structured tracing without the overhead of
    creating and binding new loggers.  It is also imperative, rather than
    functional, so we create fewer objects and so it is a little more ergonomic
    for common tracing flows.

    .. note::

        Don't create instances of this class directly — use
        :func:`~lenskit.logging.get_tracer` to create a tracer.

    Stability:
        Experimental
    """

    _logger: BoundLogger

    def __init__(self, logger: BoundLogger):
        self._logger = logger

    def add_bindings(self, **new_values: Any) -> None:
        """
        Bind new data in the keys.

        .. note::

            Unlike :meth:`structlog.Logger.bind`, this method is **imperative*: it
            updates the tracer in-place instead of returning a new tracer.  If you
            need a new, disconnected tracer, use :meth:`split`.
        """
        pass

    def remove_bindings(self, *keys: str) -> None:
        """
        Unbind keys in the tracer.

        .. note::

            Unlike :meth:`structlog.Logger.bind`, this method is **imperative*: it
            updates the tracer in-place instead of returning a new tracer.  If you
            need a new, disconnected tracer, use :meth:`split`.
        """
        pass

    def reset(self) -> None:
        """
        Reset this tracer's underlying logger to the original logger.
        """
        pass

    def trace(self, event: str, *args: Any, **bindings: Any) -> None:
        """
        Emit a TRACE-level event.
        """
        pass


class ActiveTracer(Tracer):
    """
    Active tracer that actually sends trace messages.
    """

    _base_logger: BoundLogger

    def __init__(self, logger: BoundLogger):
        super().__init__(logger)
        self._base_logger = logger

    def add_bindings(self, **new_values: Any) -> None:
        """
        Bind new data in the keys.

        .. note::

            Unlike :meth:`structlog.Logger.bind`, this method is **imperative*: it
            updates the tracer in-place instead of returning a new tracer.  If you
            need a new, disconnected tracer, use :meth:`split`.
        """
        self._logger = self._logger.bind(**new_values)

    def remove_bindings(self, *keys: str) -> None:
        """
        Unbind keys in the tracer.

        .. note::

            Unlike :meth:`structlog.Logger.bind`, this method is **imperative*: it
            updates the tracer in-place instead of returning a new tracer.  If you
            need a new, disconnected tracer, use :meth:`split`.
        """
        self._logger = self._logger.unbind(*keys)

    def reset(self) -> None:
        """
        Reset this tracer's underlying logger to the original logger.
        """
        self._logger = self._base_logger

    def trace(self, event: str, *args: Any, **bindings: Any):
        """
        Emit a TRACE-level event.
        """
        trace(self._logger, event, *args, **bindings)


class TracingLogger(structlog.stdlib.BoundLogger):
    """
    Class for LensKit loggers with trace-level logging support.

    Code should not directly use the tracing logger — it should use the
    :func:`trace` function that intelligently checks the logger.
    """

    def bind(self, **new_values: Any) -> TracingLogger:
        return super().bind(**new_values)  # type: ignore[return-value]

    def new(self, **new_values: Any) -> TracingLogger:
        return super().new(**new_values)  # type: ignore[return-value]

    def trace(self, event: str | None, *args: Any, **kw: Any):
        if args:
            kw["positional_args"] = args
        try:
            args, kwargs = self._process_event("trace", event, kw)  # type: ignore
        except structlog.DropEvent:
            return None
        self._logger.debug(*args, **kwargs)


def lenskit_filtering_logger(level: int):
    """
    No-op filtering logger.
    """

    if level < logging.DEBUG:
        return TracingLogger

    def trace(self, event: str | None, *args: Any, **kw: Any):
        pass

    base = structlog.make_filtering_bound_logger(level)
    return type(f"LensKitLoggerFilter{level}", (base,), {"trace": trace})
